import numpy as np
from cykhash.khashsets cimport Int64Set_from_buffer
from cykhash import isin_int64
from cpython cimport array


class Solver:
    """Solver is used to toggle between exclude / keep checks applied in data filter."""
    def __init__(self, direction):

        if direction == "exclude":
            self.solver = self.isin_check
        elif direction == "keep":
            self.solver = self.notin_check
        else:
            raise ValueError("filter type should be 'keep' or 'exclude'")

    def isin_check(self, value, container):
        if value in container:
            return True
        return False

    def notin_check(self, value, container):
        if value not in container:
            return True
        return False

    def check(self, value, container):
        return self.solver(value, container)


cdef has_osm_data_type(osm_data_types, tag_keys):
    cdef str key, osm_data_type
    cdef int i, n = len(tag_keys)
    for i in range(0, n):
        key = tag_keys[i]
        for osm_data_type in osm_data_types:
            if osm_data_type == key:
                return True
    return False

cdef way_is_part_of_relation(way_record, lookup_dict):
    try:
        lookup_dict[way_record["id"]]
        return True
    except KeyError as e:
        return False
    except Exception as e:
        raise e

cdef filter_osm_records(data_records,
                        data_filter,
                        osm_data_type,
                        relation_way_ids,
                        filter_type):
    """
    Filter OSM data records by given OSM tag key:value pairs.
    
    Parameters
    ----------
    data_records : list:
        A list of OSM data records. 
        A single record is a dictionary with OSM data, such as: 
           - {"highway": "primary", "maxspeed": 80, "name": "Highway-name-foo", 
              "nodes": [1111,2222,3333,4444]}
              
    data_filter : dict ( {"osm-key": ["list-of-osm-values"]} )
        A dictionary of tag-keys and associated values that will be used to
        filter the OSM data records, e.g. {"building": ["residential"]} filters
        the records where "building" key has a value "residential". 
        The records will be kept or excluded according the <filter_type> parameter.   
              
    osm_data_type : str | list
        Basic data type(s) used for filtering, such as:
         - 'highway' (for roads), 
         - 'building' (for buildings), 
         - 'landuse' (for landuse) etc.
         - ['amenity', 'shop', 'craft'] (a combination that can be useful for parsing POIs)
         
    relation_way_ids : list (optional)
        A list of way ids that belong to relations. Ways that match with these ids are always kept.
         
    filter_type : str ( 'keep' | 'exclude' ) 
        Whether the given data_filter should 'keep' or 'exclude' the records 
        where given tag:value pair is present in the record.  
    """
    cdef str rec_value
    cdef int i, N = len(data_records)

    solver = Solver(filter_type)
    filtered_data = []

    if not isinstance(osm_data_type, list):
        osm_data_type = [osm_data_type]

    if data_filter is not None:
        if len(data_filter) == 0:
            data_filter = None

    if data_filter is not None:
        # Check if there are duplicate filter values
        # e.g. a situation where {"route": "tram"} and {"railway": "tram"}
        # filters are present simultaneously.
        overlapping_filter = False

        filter_values = []
        way_filter = {}
        for key, vals in data_filter.items():

            # Check for {osm-key: True} cases
            if vals is True or vals == [True]:
                continue

            filter_values += vals
            way_filter[key] = vals

        # Update data filter
        if len(way_filter) == 0:
            data_filter = None
        else:
            data_filter = {k: v for k, v in way_filter.items()}
            filter_keys = list(data_filter.keys())

        # Check for overlapping filter
        if len(filter_values) > len(list(set(filter_values))):
            overlapping_filter = True

    relation_check = False
    if relation_way_ids is not None:
        relation_way_lookup = dict.fromkeys(relation_way_ids)
        relation_check = True

    for i in range(0, N):
        record = data_records[i]
        record_keys = list(record.keys())
        # If way is part of relation it should be kept
        # (ways that are part of relation might not have any tags)
        if relation_check:
            if way_is_part_of_relation(record, relation_way_lookup):
                filtered_data.append(record)
                continue

        if not has_osm_data_type(osm_data_type, record_keys):
            continue

        # Check if should be filtered based on given data_filter
        if data_filter is not None:
            filter_out = False
            filter_was_in_record = False

            for k, v in record.items():
                if k in filter_keys:
                    filter_was_in_record = True
                    if solver.check(v, data_filter[k]):
                        filter_out = True
                        # If there are identical filter criteria used in multiple OSM-keys
                        # Check that none of them matches, hence do not break the loop
                        # after the first match is found.
                        if not overlapping_filter:
                            break
                    else:
                        filter_out = False
                        break

            # If none of the filter keys are present in the element,
            # it should not be kept
            if not filter_was_in_record:
                continue

            if not filter_out:
                filtered_data.append(record)

        # If data_filter has not been specified,
        # all data for specified osm-keys should be kept.
        else:
            filtered_data.append(record)
    return filtered_data

cpdef _filter_array_dict_by_indices_or_mask(array_dict, indices):
    return filter_array_dict_by_indices_or_mask(array_dict, indices)

cdef filter_array_dict_by_indices_or_mask(array_dict, indices):
    return {k: v[indices] for k, v in array_dict.items()}

cdef get_lookup_khash_for_int64(int64_id_array):
    return Int64Set_from_buffer(
        memoryview(
            int64_id_array
        )
    )

cdef get_nodeid_lookup_khash(nodes):
    return get_lookup_khash_for_int64(
        np.concatenate([group['id'].tolist()
                        for group in nodes]).astype(np.int64)
    )

cdef nodes_for_way_exist_khash(nodes, node_lookup):
    v = array.array('q', nodes)
    cdef int n = len(v)

    result = array.array('B', [False] * n)
    isin_int64(v, node_lookup, result)
    if True in result:
        return True
    return False

cdef record_should_be_kept(tag, osm_keys, data_filter, filter_type):
    if tag is None:
        return False

    cdef str k, osm_key
    filter_keys = list(data_filter.keys())
    tag_keys = list(tag.keys())

    # Check if OSM key exist for the given element
    osm_key_was_found = False
    for osm_key in osm_keys:
        if osm_key in tag_keys:
            osm_key_was_found = True

    # If not, the element shouldn't be kept
    if not osm_key_was_found:
        return False

    # If there is no filter but the element is correct kind
    if len(filter_keys) == 0:
        if filter_type == "keep":
            return True
        else:
            return False

    # If there is a filter, check if match is found
    for k, v in data_filter.items():
        if k in tag_keys:
            # Check match with data filter
            if tag[k] in v:
                if filter_type == "keep":
                    return True
                else:
                    return False
            # If filter is not defined, check for 'osm_key': True
            elif v == [True] or v == True:
                if filter_type == "keep":
                    return True
                else:
                    return False

    if filter_type == "keep":
        return False
    return True

cdef filter_relation_indices(relations, osm_keys, data_filter, filter_type):
    cdef int i, n = len(relations["tags"])
    indices = []

    # Ensure keys
    if len(data_filter) == 0:
        relation_filter = {}
    else:
        relation_filter = {key: value for key, value in data_filter.items()}

    for i in range(0, n):
        tag = relations["tags"][i]
        if record_should_be_kept(tag, osm_keys, relation_filter, filter_type):
            indices.append(i)
    return indices

cdef filter_node_indices(node_arrays, osm_keys, data_filter, filter_type):
    cdef int i, n = len(node_arrays["tags"])
    indices = []

    if len(data_filter) == 0:
        node_filter = {}
    else:
        node_filter = {key: value for key, value in data_filter.items()}

    for i in range(0, n):
        tag = node_arrays["tags"][i]
        if record_should_be_kept(tag, osm_keys, node_filter, filter_type):
            indices.append(i)

    return indices


from pyrosm.data_filter cimport filter_array_dict_by_indices_or_mask
from pyrosm.geometry cimport pygeos_to_shapely, create_relation_geometry, fix_geometry
from pyrosm._arrays cimport convert_to_arrays_and_drop_empty, convert_way_records_to_lists
from pygeos import multipolygons, multilinestrings
from pygeos.predicates import is_valid
import numpy as np
import geopandas as gpd

cpdef _get_ways_for_relation(member_ids, member_roles, building_relation_ways):
    return get_ways_for_relation(member_ids, member_roles, building_relation_ways)

cdef get_ways_for_relation(member_ids, member_roles, building_relation_ways):
    # Ensure there are no duplicate member_ids/roles
    member_ids, idx = np.unique(member_ids, return_index=True)
    member_roles = member_roles[idx]

    mask = np.isin(building_relation_ways["id"], member_ids)

    # If data for relation is not available, skip
    if np.sum(mask) == 0:
        return None, None, None

    # If some of the ways were missing, use the ones that are available.
    # Might cause incorrectly shaped geometry but it's better than not having
    # the data for relation at all
    elif len(member_ids) != np.sum(mask):
        # Filter member_ids and roles accordingly
        member_mask = np.isin(member_ids, building_relation_ways["id"])
        member_ids = member_ids[member_mask]
        member_roles = member_roles[member_mask]

    ways = filter_array_dict_by_indices_or_mask(building_relation_ways, mask)
    return member_ids, member_roles, ways

cdef get_relations(relations, relation_ways, node_coordinates):
    cdef int i, j, n2, m_cnt, n = len(relations["id"])

    prepared_relations = []
    for i in range(0, n):
        rel = filter_array_dict_by_indices_or_mask(relations, [i])
        tag_keys = list(rel["tags"][0].keys())
        tag = rel["tags"][0]

        # Check if geometry should NOT be polygon
        force_linestring = False

        # ==========================================
        # Determine if relation should be LineString
        # ==========================================
        # OSM Relation keys that are typically LineStrings
        linestring_keys = ["barrier", "route",
                           "railway", "highway",
                           "waterway"]
        for lsk in linestring_keys:
            if lsk in tag_keys:
                # Railway
                # -------
                if lsk == "railway":
                    # https://wiki.openstreetmap.org/wiki/Key:railway
                    if tag["railway"] not in ["platform", "station", "turntable",
                                              "roundhouse", "traverser", "wash"]:
                        force_linestring = True
                        break
                # Highway
                # -------
                elif lsk == "highway":
                    # https://wiki.openstreetmap.org/wiki/Key:highway
                    if tag["highway"] == "pedestrian":
                        if "area" in tag_keys:
                            # If highway is pedestrian area, it should be a Polygon
                            if tag["area"] != "yes":
                                force_linestring = True
                            break
                    elif tag["highway"] not in ["platform", "rest_area", "services"]:
                        force_linestring = True
                        break
                # Waterway
                # --------
                elif lsk == "waterway":
                    # https://wiki.openstreetmap.org/wiki/Key:waterway
                    if tag["waterway"] not in ["riverbank", "dock", "boatyard",
                                               "dam", "fuel"]:
                        force_linestring = True
                        break
                else:
                    force_linestring = True
                    break

        if "area" in tag_keys:
            if tag["area"] == "no":
                force_linestring = True

        # =========================================================
        # Determine if element should be made as MultiPolygon
        # Notice: If 'force_linestring = True', this doesn't apply
        # =========================================================
        make_multipolygon = False
        if "type" in tag_keys:
            if tag["type"] in ["multipolygon", "public_transport",
                               "site", "cluster"]:
                make_multipolygon = True

        member_ids = rel["members"][0]["member_id"]
        member_roles = rel["members"][0]["member_role"]

        # Get ways for given relation
        member_ids, member_roles, ways = get_ways_for_relation(member_ids,
                                                               member_roles,
                                                               relation_ways)

        if ways is None:
            continue

        geometry = create_relation_geometry(node_coordinates,
                                            ways,
                                            member_roles,
                                            force_linestring,
                                            make_multipolygon
                                            )
        if geometry is None:
            continue

        # Check for multigeometries
        if isinstance(geometry[0], list):
            if len(geometry[0]) > 1:
                if not force_linestring:
                    geometry = multipolygons(geometry)
                else:
                    geometry = multilinestrings(geometry)
        else:
            geometry = geometry[0]

        # Check if geometry is valid
        is_valid_geom = is_valid(geometry)

        # Convert to shapely
        geometry = pygeos_to_shapely(geometry)

        # If geometry was invalid polygon try to fix
        if not force_linestring:
            if not is_valid_geom:
                geometry = fix_geometry(geometry)

        relation = dict(
            id=rel["id"][0],
            version=rel["version"][0],
            changeset=rel["changeset"][0],
            timestamp=rel["timestamp"][0],
            geometry=geometry
        )

        # Add tags
        for k, v in rel["tags"][0].items():
            relation[k] = v

        prepared_relations.append(relation)

    return prepared_relations

cdef _prepare_relations(relations, relation_ways, node_coordinates, tags_to_keep):
    # Tags to keep as separate columns
    tags_to_keep += ["id", "nodes", "timestamp", "changeset", "version", "geometry"]

    # Also geometries are parsed in this step
    relation_records = get_relations(relations, relation_ways, node_coordinates)

    # Return empty frame if no relation records were successfully parsed
    if len(relation_records) == 0:
        return gpd.GeoDataFrame()

    data = convert_way_records_to_lists(relation_records, tags_to_keep)
    arrays = convert_to_arrays_and_drop_empty(data)
    return arrays

cpdef prepare_relations(relations, relation_ways, node_coordinates, tags_to_keep):
    return _prepare_relations(relations, relation_ways, node_coordinates, tags_to_keep)



from pyrosm.config.default_tags import *
from pyrosm.config.osm_filters import get_osm_filter


class NetworkFilter:
    driving = get_osm_filter("driving")
    driving_psv = get_osm_filter("driving+psv")
    walking = get_osm_filter("walking")
    cycling = get_osm_filter("cycling")


class Tags:
    # Tags object contains configuration about the default
    # tag combinations (key:value) that will be kept as columns
    # in the resulting GeoDataFrame. All other possible tags are
    # inserted into a JSON that is stored in "tags" column.
    # These follow more or less OSM Wiki documentation:
    # https://wiki.openstreetmap.org/wiki/Map_Features
    available = ["aerialway", "aeroway", "amenity", "boundary", "building", "craft",
                 "emergency", "geological", "highway", "historic", "landuse",
                 "leisure", "natural", "office", "power", "public_transport",
                 "railway", "route", "place", "shop", "tourism", "waterway"]

    aerialway = aerialway_columns
    aeroway = aeroway_columns
    amenity = amenity_columns
    boundary = boundary_columns
    building = building_columns
    craft = craft_columns
    emergency = emergency_columns
    geological = geological_columns
    highway = highway_columns
    historic = historic_columns
    landuse = landuse_columns
    leisure = leisure_columns
    natural = natural_columns
    office = office_columns
    power = power_columns
    public_transport = public_transport_columns
    railway = railway_columns
    route = route_columns
    place = place_columns
    shop = shop_columns
    tourism = tourism_columns
    waterway = waterway_columns


class Conf:
    network_filters = NetworkFilter()
    tags = Tags()



import os
from pyrosm.utils.download import download
from pyrosm.data.geofabrik import Africa, Antarctica, Asia, AustraliaOceania, \
    Europe, NorthAmerica, SouthAmerica, CentralAmerica, Brazil, Canada, France, \
    Germany, GreatBritain, Italy, Japan, Netherlands, Poland, Russia, USA, SubRegions
from pyrosm.data.bbbike import Cities
import warnings

__all__ = ["available", "get_data", "get_path"]
_module_path = os.path.dirname(__file__)
_package_files = {"test_pbf": "test.osm.pbf", "helsinki_pbf": "Helsinki.osm.pbf"}

# Static test data
_helsinki_region_pbf = {"name": "Helsinki_region.osm.pbf",
                        "url": "https://gist.github.com/HTenkanen/"
                               "02dcfce32d447e65024d93d39ddb1812/"
                               "raw/5fe7ffb625f091591d8c29128a9e3b37870a5012/"
                               "Helsinki_region.osm.pbf"}

class DataSources:
    def __init__(self):
        self.africa = Africa()
        self.antarctica = Antarctica()
        self.asia = Asia()
        self.australia_oceania = AustraliaOceania()
        self.europe = Europe()
        self.north_america = NorthAmerica()
        self.south_america = SouthAmerica()
        self.central_america = CentralAmerica()

        self.cities = Cities()
        self.subregions = SubRegions()

        self.available = {
            "africa": self.africa.available,
            "antarctica": self.antarctica.available,
            "asia": self.asia.available,
            "australia_oceania": self.australia_oceania.available,
            "central_america": self.central_america.available,
            "europe": self.europe.available,
            "north_america": self.north_america.available,
            "south_america": self.south_america.available,
            "cities": self.cities.available,
            "subregions": self.subregions.available,
        }

        # Gather all data sources
        # Keep hidden to avoid encouraging iteration of the whole
        # world at once which most likely would end up
        # in memory error / filling the disk etc.
        self._all_sources = [k for k in self.available.keys()
                             if k not in ["cities", "subregions"]]

        for source, available in self.available.items():
            self._all_sources += available

        for subregion in self.subregions.available:
            self._all_sources += self.subregions.__dict__[subregion].available

        self._all_sources = [src.lower() for src in self._all_sources]

        # Static data for Helsinki Region
        # that should be able to download
        # (needed for tests)
        self._all_sources += ["helsinki_region_pbf"]
        self._all_sources = list(set(self._all_sources))


# Initialize DataSources
sources = DataSources()

available = {
    "test_data": list(_package_files.keys()) + ["helsinki_region_pbf"],
    "regions": {k: v for k, v in sources.available.items() if k not in ["cities", "subregions"]},
    "subregions": sources.subregions.available,
    "cities": sources.cities.available,
}


def retrieve(data, update, directory):
    return download(url=data["url"],
                    filename=data["name"],
                    update=update,
                    target_dir=directory
                    )


def search_source(name):
    for source, available in sources.available.items():
        # Cities are kept as CamelCase, so need to make lower
        if source == "cities":
            available = [src.lower() for src in available]
        if isinstance(available, list):
            if name in available:
                return sources.__dict__[source].__dict__[name]
        elif isinstance(available, dict):
            # Sub-regions should be looked one level further down
            for subregion, available2 in available.items():
                if name in available2:
                    return sources.subregions.__dict__[subregion].__dict__[name]
    raise ValueError(f"Could not retrieve url for '{name}'.")


def get_data(dataset, update=False, directory=None):
    """
    Get the path to a PBF data file, and download the data if needed.

    Parameters
    ----------
    dataset : str
        The name of the dataset. See ``pyrosm.data.available`` for
        all options.

    update : bool
        Whether the PBF file should be downloaded/updated if the dataset
        with the same name exists in the temp.

    directory : str (optional)
        Path to a directory where the PBF data will be downloaded.
        (does not apply for test data sets bundled with the package).
    """

    if not isinstance(dataset, str):
        raise ValueError(f"'dataset' should be text. Got {dataset}.")
    dataset = dataset.lower().strip()

    if dataset in _package_files:
        return os.path.abspath(os.path.join(_module_path, _package_files[dataset]))

    elif dataset == "helsinki_region_pbf":
        return retrieve(_helsinki_region_pbf,
                        update, directory)

    elif dataset in sources._all_sources:
        return retrieve(search_source(dataset),
                 update, directory)

    # Users might pass city names with spaces (e.g. Rio De Janeiro)
    elif dataset.replace(" ", "") in sources._all_sources:
        return retrieve(search_source(dataset.replace(" ", "")),
                        update, directory)

    # Users might pass country names without underscores (e.g. North America)
    elif dataset.replace(" ", "_") in sources._all_sources:
        return retrieve(search_source(dataset.replace(" ", "_")),
                        update, directory)

    # Users might pass country names with dashes instead of underscores (e.g. canary-islands)
    elif dataset.replace("-", "_") in sources._all_sources:
        return retrieve(search_source(dataset.replace("-", "_")),
                        update, directory)

    else:
        msg = "The dataset '{data}' is not available. ".format(data=dataset)
        msg += "Available datasets are {}".format(", ".join(sources._all_sources))
        raise ValueError(msg)


# Keep temporarily for backward compatibility
def get_path(dataset, update=False, directory=None):
    warnings.warn(
        "'get_path()' is deprecated, use 'get_data()' instead.",
        DeprecationWarning,
        stacklevel=2
    )
    return get_data(dataset, update, directory)



import urllib
import tempfile
import os
import enum
from urllib.error import HTTPError

class UNIT(enum.Enum):
    BYTES = 1
    KB = 2
    MB = 3
    GB = 4


def convert_unit(size_in_bytes, unit):
    if unit == UNIT.KB:
        return size_in_bytes / 1024
    elif unit == UNIT.MB:
        return size_in_bytes / (1024 * 1024)
    elif unit == UNIT.GB:
        return size_in_bytes / (1024 * 1024 * 1024)
    else:
        return size_in_bytes


def get_file_size(file_name, size_type=UNIT.MB):
    size = os.path.getsize(file_name)
    return round(convert_unit(size, size_type), 2)


def download(url, filename, update, target_dir):
    if target_dir is None:
        temp_dir = tempfile.gettempdir()
        target_dir = os.path.join(temp_dir, 'pyrosm')
    else:
        if not os.path.isdir(target_dir):
            raise ValueError(f"The provided directory does not exist: "
                             f"{target_dir}")

    filepath = os.path.abspath(os.path.join(target_dir, os.path.basename(filename)))

    if not os.path.exists(target_dir):
        os.makedirs(target_dir)

    # Check if file exists
    file_exists = False
    if os.path.exists(filepath):
        file_exists = True

    if update and file_exists:
        os.remove(filepath)

    # Download data to temp if it does not exist or if update is requested
    if update or file_exists is False:

        try:
            filepath, msg = urllib.request.urlretrieve(url, filepath)
        except HTTPError:
            raise ValueError(f"PBF-file '{url}' is temporarily unavailable. "
                             f"Try again later.")
        except Exception as e:
            raise e

        filesize = get_file_size(filepath)
        if filesize == 0:
            raise ValueError(f"PBF-file '{filename}' from the provider was empty. "
                             "This is likely a temporary issue, try again later."
                             )
        print(f"Downloaded Protobuf data '{os.path.basename(filepath)}' "
              f"({filesize} MB) to:\n'{filepath}'")
    return filepath



import numpy as np


cdef delta_decode_latitude(data, node_granularity, lat_offset):
    div = 1000000000
    lats_deltas = np.zeros(len(data.lat) + 1, dtype=np.int64)
    lats_deltas[1:] = list(data.lat)
    lats = (np.cumsum(lats_deltas)[1:] * node_granularity + lat_offset) / div
    return lats


cdef delta_decode_longitude(data, node_granularity, lon_offset):
    div = 1000000000
    lons_deltas = np.zeros(len(data.lon) + 1, dtype=np.int64)
    lons_deltas[1:] = list(data.lon)
    lons = (np.cumsum(lons_deltas)[1:] * node_granularity + lon_offset) / div
    return lons


cdef delta_decode_id(data):
    id_deltas = np.zeros(len(data.id) + 1, dtype=np.int64)
    id_deltas[1:] = list(data.id)
    ids = np.cumsum(id_deltas)[1:]
    return ids


cdef delta_decode_timestamp(data, timestamp_granularity):
    timestamp_deltas = np.zeros(len(data.denseinfo.timestamp) + 1, dtype=np.int64)
    timestamp_deltas[1:] = list(data.denseinfo.timestamp)
    timestamps = (np.cumsum(timestamp_deltas)[1:]
                  * timestamp_granularity / 1000).astype(int)
    return timestamps


cdef delta_decode_changeset(data):
    changeset_deltas = np.zeros(len(data.denseinfo.changeset) + 1, dtype=np.int64)
    changeset_deltas[1:] = list(data.denseinfo.changeset)
    changesets = np.cumsum(changeset_deltas)[1:]
    return changesets


cdef delta_encode_latitude(lat_array, node_granularity, lat_offset):
    # Note: There might be small difference in the coordinates (1 cm) after
    # encoding due to rounding errors
    div = 1000000000
    lat_deltas = np.zeros(len(lat_array) + 1, dtype=np.float64)
    lat_deltas[1:] = list(lat_array)
    lat_deltas = (np.diff(lat_deltas) / node_granularity - lat_offset) * div
    lat_deltas = lat_deltas.astype(np.int64)
    return lat_deltas


cdef delta_encode_longitude(lon_array, node_granularity, lon_offset):
    # Note: There might be small difference in the coordinates (1 cm) after
    # encoding due to rounding errors
    div = 1000000000
    lon_deltas = np.zeros(len(lon_array) + 1, dtype=np.float64)
    lon_deltas[1:] = list(lon_array)
    lon_deltas = (np.diff(lon_deltas) / node_granularity - lon_offset) * div
    lon_deltas = lon_deltas.astype(np.int64)
    return lon_deltas

cdef delta_encode_id(id_array):
    id_deltas = np.zeros(len(id_array) + 1, dtype=np.int64)
    id_deltas[1:] = list(id_array)
    ids = np.diff(id_deltas)
    return ids


cpdef create_nodes_gdf(node_dict_list)
cpdef create_gdf(data_records, geometry_array)
cpdef prepare_way_gdf(node_coordinates, ways)
cpdef prepare_node_gdf(nodes)
cpdef prepare_geodataframe(nodes, node_coordinates, ways,
                           relations, relation_ways, tags_as_columns,
                           bounding_box)


from libc.stdlib cimport malloc
import cython
import numpy as np
from rapidjson import dumps
from geopandas import GeoSeries

cdef get_dtype(key):
    dtypes = {"id": np.int64,
              "version": np.int8,
              "changeset": np.int8,
              "timestamp": np.int64,
              "lon": np.float32,
              "lat": np.float32,
              "tags": object,
              "members": object,
              "geometry": object,
              }
    if key in dtypes.keys():
        return dtypes[key]
    return None

cdef convert_way_records_to_lists(ways, tags_to_separate_as_arrays):
    """
    Function to convert heterogeneous way dictionaries into harmonized dictionary 
    of value-lists for all OSM keys. If a given OSM-key is not present for a given way 
    record, will add None. This process makes it possible to create same-sized numpy arrays
    for all OSM way tags.  
    """
    cdef int i
    cdef int n = len(ways)

    lookup = dict.fromkeys(tags_to_separate_as_arrays, None)
    data = {k: [] for k in tags_to_separate_as_arrays}
    data["tags"] = []

    for i in range(0, n):
        way = ways[i]
        way_records = dict.fromkeys(tags_to_separate_as_arrays, None)
        other_tags = {}
        for k, v in way.items():
            try:
                # Check if tag should be kept as a column
                lookup[k]
                way_records[k] = v
            except:
                # If not add into tags
                other_tags[k] = v
        [data[k].append(v) for k, v in way_records.items()]
        if len(other_tags) > 0:
            data["tags"].append(dumps(other_tags))
        else:
            data["tags"].append(None)

    return data

cdef convert_to_arrays_and_drop_empty(data):
    """
    Function to convert the harmonized dictionary 
    of key:values to numpy arrays. If a given OSM-key contains
    only None values (i.e. is "empty"), it will be dropped from the 
    resulting dictionary. 
    """
    # Convert to arrays
    arrays = {}
    for key, value_list in data.items():
        # Parse geometry separately for handling multi-geoms correctly
        if key == "geometry":
            arrays[key] = GeoSeries(value_list).values
            continue

        # Nodes are in a list and should always be kept
        elif not isinstance(value_list[0], list):
            # Keep tag only if it contains data
            unique = list(set(value_list))
            if len(unique) < 2:
                if unique[0] is None:
                    continue
        try:
            arrays[key] = np.array(value_list, dtype=get_dtype(key))
        except ValueError as e:
            if "invalid literal for int" in str(e):
                # Try first converting to int via floats
                try:
                    value_list = list(map(int, list(map(float, value_list))))
                    arrays[key] = np.array(value_list, dtype=np.int64)
                except ValueError as e:
                    # If there is a string, keep as is
                    if "convert string" in str(e):
                        arrays[key] = np.array(value_list, dtype=object)
                except Exception as e:
                    raise e
        except Exception as e:
            raise e

    return arrays

cpdef concatenate_dicts_of_arrays(dict_list_of_arrays):
    cdef str k

    keys = list(set([k for d in dict_list_of_arrays
                     for k in d.keys()]))
    result_dict = {key: [] for key in keys}

    for dicts in dict_list_of_arrays:
        for k, v in dicts.items():
            result_dict[k] += v.tolist()

    # Convert to arrays
    result_arrays = {}
    for k, v in result_dict.items():
        if len(v) > 0:
            result_arrays[k] = np.array(v, dtype=get_dtype(k))

    # The length of all arrays must match
    length = None
    for k, array in result_arrays.items():
        arr_cnt = array.shape[0]
        if length is None:
            length = arr_cnt
        else:
            assert length == arr_cnt, f"The length of '{k}' " \
                                      f"should be {length}, " \
                                      f"got {array.shape[0]}."

    return result_arrays

cdef char** to_cstring_array(list str_list):
    """
    Converts Python byte-string list to an "array" of c-strings. 
    NOTE: Memory handling needs to be done manually in the main application!
    """
    cdef int i, N = len(str_list)
    cdef char ** string_array = <char **> malloc(N * sizeof(char *))
    cdef char *txt

    if not string_array:
        raise MemoryError()

    for i in range(0, N):
        txt = str_list[i]
        string_array[i] = txt

    return string_array

cdef int*to_cint_array(list int_list):
    """
    Converts Python list of integers to an "array" of C-integers. 
    NOTE: Memory handling needs to be done manually in the main application!
    """
    cdef int *c_ints
    cdef int N = len(int_list)

    c_ints = <int *> malloc(N * cython.sizeof(int))

    if not c_ints:
        raise MemoryError()

    for i in range(0, N):
        c_ints[i] = int_list[i]

    return c_ints

cdef float*to_cfloat_array(list float_list):
    """
    Converts Python list of floats to an "array" of C-floats. 
    NOTE: Memory handling needs to be done manually in the main application!
    """
    cdef float *c_floats
    cdef int N = len(float_list)

    c_floats = <float *> malloc(N * cython.sizeof(float))

    if not c_floats:
        raise MemoryError()

    for i in range(0, N):
        c_floats[i] = float_list[i]

    return c_floats

cdef long long*to_clong_array(long_list):
    """
    Converts Python list of integers (long) to an "array" of C-long longs. 
    NOTE: Memory handling needs to be done manually in the main application!
    """

    cdef long long *c_longs
    cdef int N = len(long_list)

    c_longs = <long long *> malloc(N * sizeof(long long))

    if not c_longs:
        raise MemoryError()

    for i in range(0, N):
        c_longs[i] = long_list[i]

    return c_longs



from pyrosm.config import Conf
from pyrosm.pbfreader import parse_osm_data
from pyrosm._arrays import concatenate_dicts_of_arrays
from pyrosm.geometry import create_node_coordinates_lookup
from pyrosm.frames import create_nodes_gdf
from pyrosm.utils import validate_custom_filter, validate_osm_keys, \
    validate_tags_as_columns, validate_booleans, validate_boundary_type, \
    validate_bounding_box, validate_input_file, get_bounding_box
from pyrosm.utils.download import get_file_size
from shapely.geometry import Polygon, MultiPolygon, \
    LineString, LinearRing, MultiLineString
from pyrosm.boundary import get_boundary_data
from pyrosm.buildings import get_building_data
from pyrosm.landuse import get_landuse_data
from pyrosm.natural import get_natural_data
from pyrosm.networks import get_network_data
from pyrosm.pois import get_poi_data
from pyrosm.user_defined import get_user_defined_data


class OSM:
    from pyrosm.utils._compat import PYGEOS_SHAPELY_COMPAT
    allowed_bbox_types = [Polygon, MultiPolygon, MultiLineString,
                          LineString, LinearRing]

    def __init__(self, filepath,
                 bounding_box=None):
        """
        Parameters
        ----------

        filepath : str
            Filepath to input OSM dataset ( *.osm.pbf )

        bounding_box : list | shapely geometry
            Filtering OSM data spatially is allowed by passing a
            bounding box either as a list `[minx, miny, maxx, maxy]` or
            as a Shapely Polygon/MultiPolygon or closed LineString/LinearRing.

        """

        # Check input file
        self.filepath = validate_input_file(filepath)

        if bounding_box is None:
            self.bounding_box = None
        elif type(bounding_box) in self.allowed_bbox_types:
            # Ensures bounding box is a closed geometry
            # (+ attempts to close MultiLineStrings)
            self.bounding_box = validate_bounding_box(bounding_box)
        elif isinstance(bounding_box, list):
            if not len(bounding_box) == 4:
                raise ValueError("When passing bounding box as a list it should contain 4 coordinates: "
                                 "[minx, miny, maxx, maxy].")
            self.bounding_box = bounding_box
        else:
            raise ValueError("bounding_box should be a list, Shapely Polygon or a Shapely LinearRing.")

        self.conf = Conf
        self.keep_node_info = False

        # Update file size
        self.file_size = get_file_size(self.filepath)

        # Get bounding box
        self._data_bounding_box = get_bounding_box(self.filepath)

        # TODO: Add logging as a parameter?
        self._verbose = False

        self._all_way_tags = None
        self._nodes = None
        self._nodes_gdf = None
        self._node_coordinates = None
        self._way_records = None
        self._relations = None

    def _read_pbf(self):
        # PBF reading requires a list of bounding box coordinates
        if type(self.bounding_box) in self.allowed_bbox_types:
            bounding_box = self.bounding_box.bounds
        else:
            bounding_box = self.bounding_box

        nodes, ways, relations, way_tags = parse_osm_data(self.filepath,
                                             bounding_box,
                                             exclude_relations=False)

        self._nodes = nodes
        self._way_records = ways
        self._relations = relations
        self._all_way_tags = way_tags

        # Prepare node coordinates lookup table
        self._node_coordinates = create_node_coordinates_lookup(self._nodes)

    def _get_network_filter(self, net_type):
        possible_filters = [a for a in self.conf.network_filters.__dir__()
                            if "__" not in a]
        possible_filters += ["all", "driving+service"]
        possible_values = ", ".join(possible_filters)
        msg = "'network_type' should be one of the following: " + possible_values
        if not isinstance(net_type, str):
            raise ValueError(msg)

        net_type = net_type.lower()

        if net_type not in possible_filters:
            raise ValueError(msg)

        # Get filter
        if net_type == "walking":
            return self.conf.network_filters.walking
        elif net_type == "driving":
            return self.conf.network_filters.driving
        elif net_type == "driving+service":
            return self.conf.network_filters.driving_psv
        elif net_type == "cycling":
            return self.conf.network_filters.cycling
        elif net_type == "all":
            return None

    def get_network(self, network_type="walking", extra_attributes=None):
        """
        Parses street networks from OSM
        for walking, driving, and cycling.

        Parameters
        ----------

        network_type : str
            What kind of network to parse.
            Possible values are:
              - `'walking'`
              - `'cycling'`
              - `'driving'`
              - `'driving+service'`
              - `'all'`.

        extra_attributes : list (optional)
            Additional OSM tag keys that will be converted into columns in the resulting GeoDataFrame.

        See Also
        --------

        Take a look at the OSM documentation for further details about the data:
        `https://wiki.openstreetmap.org/wiki/Key:highway <https://wiki.openstreetmap.org/wiki/Key:highway>`__

        """
        # Get filter
        network_filter = self._get_network_filter(network_type)
        tags_as_columns = self.conf.tags.highway

        if extra_attributes is not None:
            validate_tags_as_columns(extra_attributes)
            tags_as_columns += extra_attributes

        if self._nodes is None or self._way_records is None:
            self._read_pbf()

        # Filter network data with given filter
        gdf = get_network_data(self._node_coordinates,
                               self._way_records,
                               tags_as_columns,
                               network_filter,
                               self.bounding_box
                               )

        # Do not keep node information unless specifically asked for
        # (they are in a list, and can cause issues when saving the files)
        if not self.keep_node_info and gdf is not None:
            if "nodes" in gdf.columns:
                gdf = gdf.drop("nodes", axis=1)

        return gdf

    def get_buildings(self, custom_filter=None, extra_attributes=None):
        """
        Parses buildings from OSM.

        Parameters
        ----------

        custom_filter : dict
            What kind of buildings to parse,
            see details below.

            You can opt-in specific elements by using 'custom_filter'.
            To keep only specific buildings such as 'residential' and 'retail', you can apply
            a custom filter which is a Python dictionary with following format:
              - `custom_filter={'building': ['residential', 'retail']}`

        extra_attributes : list (optional)
            Additional OSM tag keys that will be converted into columns in the resulting GeoDataFrame.

        See Also
        --------

        Take a look at the OSM documentation for further details about the data:
        `https://wiki.openstreetmap.org/wiki/Key:building <https://wiki.openstreetmap.org/wiki/Key:building>`__

        """
        # Default tags to keep as columns
        tags_as_columns = self.conf.tags.building

        if extra_attributes is not None:
            validate_tags_as_columns(extra_attributes)
            tags_as_columns += extra_attributes

        if self._nodes is None or self._way_records is None:
            self._read_pbf()

        gdf = get_building_data(self._node_coordinates,
                                self._way_records,
                                self._relations,
                                tags_as_columns,
                                custom_filter,
                                self.bounding_box
                                )

        # Do not keep node information unless specifically asked for
        # (they are in a list, and can cause issues when saving the files)
        if not self.keep_node_info and gdf is not None:
            if "nodes" in gdf.columns:
                gdf = gdf.drop("nodes", axis=1)
        return gdf

    def get_landuse(self, custom_filter=None, extra_attributes=None):
        """
        Parses landuse from OSM.

        Parameters
        ----------

        custom_filter : dict
            What kind of landuse to parse,
            see details below.

            You can opt-in specific elements by using 'custom_filter'.
            To keep only specific landuse such as 'construction' and 'industrial', you can apply
            a custom filter which is a Python dictionary with following format:
              `custom_filter={'landuse': ['construction', 'industrial']}`

        extra_attributes : list (optional)
            Additional OSM tag keys that will be converted into columns in the resulting GeoDataFrame.

        See Also
        --------

        Take a look at OSM documentation for further details about the data:

        `https://wiki.openstreetmap.org/wiki/Key:landuse <https://wiki.openstreetmap.org/wiki/Key:landuse>`__

        """

        if self._nodes is None or self._way_records is None:
            self._read_pbf()

        # Default tags to keep as columns
        tags_as_columns = self.conf.tags.landuse

        if extra_attributes is not None:
            validate_tags_as_columns(extra_attributes)
            tags_as_columns += extra_attributes

        # If nodes are still in chunks, merge before passing forward
        if isinstance(self._nodes, list):
            self._nodes = concatenate_dicts_of_arrays(self._nodes)

        gdf = get_landuse_data(self._nodes,
                               self._node_coordinates,
                               self._way_records,
                               self._relations,
                               tags_as_columns,
                               custom_filter,
                               self.bounding_box
                               )

        # Do not keep node information unless specifically asked for
        # (they are in a list, and can cause issues when saving the files)
        if not self.keep_node_info and gdf is not None:
            if "nodes" in gdf.columns:
                gdf = gdf.drop("nodes", axis=1)
        return gdf

    def get_natural(self, custom_filter=None, extra_attributes=None):
        """
        Parses natural from OSM.

        Parameters
        ----------

        custom_filter : dict
            What kind of natural to parse,
            see details below.

            You can opt-in specific elements by using 'custom_filter'.
            To keep only specific natural such as 'wood' and 'tree', you can apply
            a custom filter which is a Python dictionary with following format:
              `custom_filter={'natural': ['wood', 'tree']}`

        extra_attributes : list (optional)
            Additional OSM tag keys that will be converted into columns in the resulting GeoDataFrame.

        See Also
        --------

        Take a look at OSM documentation for further details about the data:

        `https://wiki.openstreetmap.org/wiki/Key:natural <https://wiki.openstreetmap.org/wiki/Key:natural>`__

        """

        if self._nodes is None or self._way_records is None:
            self._read_pbf()

        # Default tags to keep as columns
        tags_as_columns = self.conf.tags.natural

        if extra_attributes is not None:
            validate_tags_as_columns(extra_attributes)
            tags_as_columns += extra_attributes

        # If nodes are still in chunks, merge before passing forward
        if isinstance(self._nodes, list):
            self._nodes = concatenate_dicts_of_arrays(self._nodes)

        gdf = get_natural_data(self._nodes,
                               self._node_coordinates,
                               self._way_records,
                               self._relations,
                               tags_as_columns,
                               custom_filter,
                               self.bounding_box)

        # Do not keep node information unless specifically asked for
        # (they are in a list, and can cause issues when saving the files)
        if not self.keep_node_info and gdf is not None:
            if "nodes" in gdf.columns:
                gdf = gdf.drop("nodes", axis=1)
        return gdf

    def get_boundaries(self, boundary_type="administrative", name=None, custom_filter=None, extra_attributes=None):
        """
        Parses boundaries from OSM.

        Parameters
        ----------

        boundary_type : str
            The type of boundaries to parse. Possible values:
              - `"administrative"` (default)
              - `"national_park"`
              - `"political"`
              - `"postal_code"`
              - `"protected_area"`
              - `"aboriginal_lands"`
              - `"maritime"`
              - `"lot"`
              - `"parcel"`
              - `"tract"`
              - `"marker"`
              - `"all"`

        name : str (optional)
            Name of the administrative area that will be searched for.

        custom_filter : dict (optional)
            Additional filter for what kind of boundary to parse.

        extra_attributes : list (optional)
            Additional OSM tag keys that will be converted into columns in the resulting GeoDataFrame.


        See Also
        --------

        Take a look at OSM documentation for further details about the data:

        `https://wiki.openstreetmap.org/wiki/Key:boundary <https://wiki.openstreetmap.org/wiki/Key:bondary>`__

        """

        if self._nodes is None or self._way_records is None:
            self._read_pbf()

        # Default tags to keep as columns
        tags_as_columns = self.conf.tags.boundary

        if extra_attributes is not None:
            validate_tags_as_columns(extra_attributes)
            tags_as_columns += extra_attributes

        # If nodes are still in chunks, merge before passing forward
        if isinstance(self._nodes, list):
            self._nodes = concatenate_dicts_of_arrays(self._nodes)

        # Check boundary type
        boundary_type = validate_boundary_type(boundary_type)

        if name is not None:
            if not isinstance(name, str):
                raise ValueError(f"'name' should be text."
                                 f"Got '{name}' of type {type(name)}.")

        gdf = get_boundary_data(self._node_coordinates,
                                self._way_records,
                                self._relations,
                                tags_as_columns,
                                custom_filter,
                                boundary_type,
                                name,
                                self.bounding_box)

        # Do not keep node information unless specifically asked for
        # (they are in a list, and can cause issues when saving the files)
        if not self.keep_node_info and gdf is not None:
            if "nodes" in gdf.columns:
                gdf = gdf.drop("nodes", axis=1)
        return gdf

    def get_pois(self, custom_filter=None, extra_attributes=None):
        """
        Parse Point of Interest (POI) from OSM.

        Parameters
        ----------

        custom_filter : dict
            An optional custom filter to filter only specific POIs from OpenStreetMap,
            see details below.

        extra_attributes : list (optional)
            Additional OSM tag keys that will be converted into columns in the resulting GeoDataFrame.


        Notes
        -----

        By default, Pyrosm will parse all OSM elements (points, lines and polygons)
        that are associated with following keys:
          - amenity
          - shop
          - tourism

        You can opt-out / opt-in specific elements by using 'custom_filter'.
        To parse elements associated with only specific tags, such as amenities,
        you can specify:
          `custom_filter={"amenity": True}`

        You can also combine multiple filters at the same time.
        For instance, you can parse all 'amenity' elements AND specific 'shop' elements,
        such as supermarkets and book stores by specifying:
          `custom_filter={"amenity": True, "shop": ["supermarket", "books"]}`


        See Also
        --------

        You can check the most typical OSM tags for different map features from OSM Wiki
        `https://wiki.openstreetmap.org/wiki/Map_Features <https://wiki.openstreetmap.org/wiki/Map_Features>`__.
        It is also possible to get a quick look at the most typical OSM tags from Pyrosm configuration:

        >>> from pyrosm.config import Conf
        >>> print("All available OSM keys", Conf.tags.available)
        All available OSM keys ['aerialway', 'aeroway', 'amenity', 'building', 'craft',
        'emergency', 'geological', 'highway', 'historic', 'landuse', 'leisure',
        'natural', 'office', 'power', 'public_transport', 'railway', 'route',
        'place', 'shop', 'tourism', 'waterway']

        >>> print("Typical tags associated with tourism:", Conf.tags.tourism)
        ['alpine_hut', 'apartment', 'aquarium', 'artwork', 'attraction', 'camp_pitch',
        'camp_site', 'caravan_site', 'chalet', 'gallery', 'guest_house', 'hostel',
        'hotel', 'information', 'motel', 'museum', 'picnic_site', 'theme_park',
        'tourism', 'viewpoint', 'wilderness_hut', 'zoo']

        """
        # If custom_filter has not been defined, initialize with default
        if custom_filter is None:
            custom_filter = {"amenity": True,
                             "shop": True,
                             "tourism": True
                             }

        else:
            # Check that the custom filter is in correct format
            if not isinstance(custom_filter, dict):
                raise ValueError(f"'custom_filter' should be a Python dictionary. "
                                 f"Got {custom_filter} with type {type(custom_filter)}.")

        if self._nodes is None or self._way_records is None:
            self._read_pbf()

        # Default tags to keep as columns
        tags_as_columns = []
        for k in custom_filter.keys():
            tags_as_columns += getattr(self.conf.tags, k)

        if extra_attributes is not None:
            validate_tags_as_columns(extra_attributes)
            tags_as_columns += extra_attributes

        # If nodes are still in chunks, merge before passing forward
        if isinstance(self._nodes, list):
            self._nodes = concatenate_dicts_of_arrays(self._nodes)

        gdf = get_poi_data(self._nodes,
                           self._node_coordinates,
                           self._way_records,
                           self._relations,
                           tags_as_columns,
                           custom_filter,
                           self.bounding_box)

        # Do not keep node information unless specifically asked for
        # (they are in a list, and can cause issues when saving the files)
        if not self.keep_node_info and gdf is not None:
            if "nodes" in gdf.columns:
                gdf = gdf.drop("nodes", axis=1)
        return gdf

    def get_data_by_custom_criteria(self,
                                    custom_filter,
                                    osm_keys_to_keep=None,
                                    filter_type="keep",
                                    tags_as_columns=None,
                                    keep_nodes=True,
                                    keep_ways=True,
                                    keep_relations=True,
                                    extra_attributes=None):
        """
        Parse OSM data based on custom criteria.

        Parameters
        ----------

        custom_filter : dict (required)
            A custom filter to filter only specific POIs from OpenStreetMap.

        osm_keys_to_keep : str | list
            A filter to specify which OSM keys should be kept.

        filter_type : str
            "keep" | "exclude"
            Whether the filters should be used to keep or exclude the data from OSM.

        tags_as_columns : list
            Which tags should be kept as columns in the resulting GeoDataFrame.

        keep_nodes : bool
            Whether or not the nodes should be kept in the resulting GeoDataFrame if they are found.

        keep_ways : bool
            Whether or not the ways should be kept in the resulting GeoDataFrame if they are found.

        keep_relations : bool
            Whether or not the relations should be kept in the resulting GeoDataFrame if they are found.

        extra_attributes : list (optional)
            Additional OSM tag keys that will be converted into columns in the resulting GeoDataFrame.

        """

        # Check that the custom filter is in correct format
        custom_filter = validate_custom_filter(custom_filter)

        if not isinstance(filter_type, str):
            raise ValueError("'filter_type' -parameter should be either 'keep' or 'exclude'. ")

        # Validate osm keys
        validate_osm_keys(osm_keys_to_keep)
        if isinstance(osm_keys_to_keep, str):
            osm_keys_to_keep = [osm_keys_to_keep]

        # Validate filter
        filter_type = filter_type.lower()
        if filter_type not in ["keep", "exclude"]:
            raise ValueError("'filter_type' -parameter should be either 'keep' or 'exclude'. ")

        # Tags to keep as columns
        if tags_as_columns is None:
            tags_as_columns = []
            for k in custom_filter.keys():
                try:
                    tags_as_columns += getattr(self.conf.tags, k)
                except Exception as e:
                    pass
            # If tags weren't available in conf, store keys as columns by default
            # (all other tags in such cases will be stored in 'tags' column as JSON)
            if len(tags_as_columns) == 0:
                tags_as_columns = list(custom_filter.keys())

        else:
            # Validate tags
            validate_tags_as_columns(tags_as_columns)

        if extra_attributes is not None:
            validate_tags_as_columns(extra_attributes)
            tags_as_columns += extra_attributes

        # Validate booleans
        validate_booleans(keep_nodes, keep_ways, keep_relations)

        if self._nodes is None or self._way_records is None:
            self._read_pbf()

        # If nodes are still in chunks, merge before passing forward
        if isinstance(self._nodes, list):
            self._nodes = concatenate_dicts_of_arrays(self._nodes)

        gdf = get_user_defined_data(self._nodes,
                                    self._node_coordinates,
                                    self._way_records,
                                    self._relations,
                                    tags_as_columns,
                                    custom_filter,
                                    osm_keys_to_keep,
                                    filter_type,
                                    keep_nodes,
                                    keep_ways,
                                    keep_relations,
                                    self.bounding_box
                                    )

        # Do not keep node information unless specifically asked for
        # (they are in a list, and can cause issues when saving the files)
        if not self.keep_node_info and gdf is not None:
            if "nodes" in gdf.columns:
                gdf = gdf.drop("nodes", axis=1)
        return gdf

    def __getattribute__(self, name):
        # If node-gdf is requested convert to gdf before returning
        if name == "_nodes_gdf":
            return create_nodes_gdf(super(OSM, self).__getattribute__("_nodes"))
        return super(OSM, self).__getattribute__(name)



cpdef create_point_geometries(xarray, yarray)
cdef _create_point_geometries(xarray, yarray)
cdef _create_way_geometries(node_coordinates, way_elements)
cpdef create_way_geometries(node_coordinates, way_elements)
cdef create_relation_geometry(node_coordinates, ways,
                             member_roles, force_linestring,
                             make_multipolygon)
cdef create_linear_ring(coordinates)
cdef create_linestring_geometry(nodes, node_coordinates)
cpdef create_node_coordinates_lookup(nodes)
cdef pygeos_to_shapely(geom)
cdef to_shapely(pygeos_array)
cdef get_way_coordinates_for_polygon(node_coordinate_lookup, way_records)
cpdef fix_geometry(geometry, diff_threshold=*)



from rapidjson import dumps

cdef unicode tounicode(char *s):
    return s.decode("UTF-8")

cdef parse_tags(keys, vals, stringtable):
    cdef int i, N=len(keys)
    d = dict()
    for i in range(0, N):
        k, v = keys[i], vals[i]
        d[stringtable[k]] = stringtable[v]
    return d

cdef parse_dense_tags(keys_vals, string_table):
    cdef int N=len(keys_vals)
    cdef int tag_idx = 0
    tag_list = []

    while tag_idx < N:
        tags = dict()
        while keys_vals[tag_idx] != 0:
            k = keys_vals[tag_idx]
            v = keys_vals[tag_idx + 1]
            tag_idx += 2
            tags[string_table[k]] = string_table[v]

        if len(tags) > 0:
            tag_list.append(tags)
        else:
            tag_list.append(None)
        tag_idx += 1
    return tag_list

cdef explode_way_tags(ways):
    exploded = []
    cdef int i, n=len(ways)
    way_keys = {}
    for i in range(0, n):
        way = ways[i]
        for k, v in way['tags'].items():
            way[k] = v
            try:
                dummy = way_keys[k]
            except:
                way_keys[k] = None
        del way['tags']
        exploded.append(way)
    return exploded, list(way_keys.keys())

cdef explode_tag_array(tag_array, tags_as_columns):
    lookup = dict.fromkeys(tags_as_columns, None)
    data = {k: [] for k in tags_as_columns}
    data["tags"] = []
    n = len(tag_array)
    for i in range(0, n):
        tag = tag_array[i]
        tag_records = dict.fromkeys(tags_as_columns, None)
        other_tags = {}
        for k, v in tag.items():
            try:
                # Check if tag should be kept as a column
                lookup[k]
                tag_records[k] = v
            except:
                # If not add into tags
                other_tags[k] = v
        [data[k].append(v) for k, v in tag_records.items()]
        if len(other_tags) > 0:
            data["tags"].append(dumps(other_tags))
        else:
            data["tags"].append(None)
    return data


import pandas as pd
import geopandas as gpd
from pyrosm._arrays cimport concatenate_dicts_of_arrays
from pyrosm.geometry cimport _create_point_geometries
from pyrosm.geometry cimport create_way_geometries
from pyrosm.relations import prepare_relations
from shapely.geometry import box


cpdef create_nodes_gdf(nodes):
    cdef str k
    if isinstance(nodes, list):
        nodes = concatenate_dicts_of_arrays(nodes)
    df = pd.DataFrame()
    for k, v in nodes.items():
        df[k] = v
    df['geometry'] = _create_point_geometries(nodes['lon'], nodes['lat'])
    return gpd.GeoDataFrame(df, crs='epsg:4326')

cpdef create_gdf(data_arrays, geometry_array):
    cdef str key
    df = pd.DataFrame()
    for key, data in data_arrays.items():
        # When inserting nodes,
        # those should be converted
        # to lists to avoid block error
        if key == "nodes":
            df[key] = data.tolist()
        else:
            df[key] = data

    df['geometry'] = geometry_array
    return gpd.GeoDataFrame(df, crs='epsg:4326')

cpdef prepare_way_gdf(node_coordinates, ways):
    if ways is not None:
        geometries = create_way_geometries(node_coordinates,
                                           ways)
        # Convert to GeoDataFrame
        way_gdf = create_gdf(ways, geometries)
        way_gdf['osm_type'] = "way"
    else:
        way_gdf = gpd.GeoDataFrame()
    return way_gdf

cpdef prepare_node_gdf(nodes):
    if nodes is not None:
        # Create GeoDataFrame from nodes
        node_gdf = create_nodes_gdf(nodes)
        node_gdf['osm_type'] = "node"
    else:
        node_gdf = gpd.GeoDataFrame()
    return node_gdf

cpdef prepare_relation_gdf(node_coordinates, relations, relation_ways, tags_as_columns):
    if relations is not None:
        relations = prepare_relations(relations, relation_ways,
                                      node_coordinates,
                                      tags_as_columns)

        relation_gdf = gpd.GeoDataFrame(relations)
        relation_gdf['osm_type'] = "relation"

    else:
        relation_gdf = gpd.GeoDataFrame()
    return relation_gdf

cpdef prepare_geodataframe(nodes, node_coordinates, ways,
                           relations, relation_ways,
                           tags_as_columns, bounding_box):
    # Prepare nodes
    node_gdf = prepare_node_gdf(nodes)

    # Prepare ways
    way_gdf = prepare_way_gdf(node_coordinates, ways)

    # Prepare relation data
    relation_gdf = prepare_relation_gdf(node_coordinates, relations, relation_ways, tags_as_columns)

    # Merge all
    gdf = pd.concat([node_gdf, way_gdf, relation_gdf])

    if len(gdf) == 0:
        return None

    gdf = gdf.dropna(subset=['geometry']).reset_index(drop=True)

    # Filter by bounding box if it was used
    # This would be faster using Pygeos,
    # but as GeoPandas 0.8.0 should start supporting it natively,
    # let's keep things simple
    if bounding_box is not None:
        if isinstance(bounding_box, list):
            bounding_box = box(*bounding_box)
        # Filter data spatially
        orig_cols = list(gdf.columns)
        filter_gdf = gpd.GeoDataFrame({"geometry": [bounding_box]},
                                      crs="epsg:4326",
                                      index=[0])
        gdf = gpd.sjoin(gdf, filter_gdf, how="inner")
        gdf = gdf[orig_cols].reset_index(drop=True)

    if len(gdf) == 0:
        return None

    return gdf



from pyrosm.data_manager import get_osm_data
from pyrosm.frames import prepare_geodataframe
import warnings


def get_user_defined_data(nodes,
                          node_coordinates,
                          way_records,
                          relations,
                          tags_as_columns,
                          custom_filter,
                          osm_keys,
                          filter_type,
                          keep_nodes,
                          keep_ways,
                          keep_relations,
                          bounding_box):

    if not keep_nodes:
        nodes = None

    # If wanting to parse relations but not ways,
    # it is still necessary to parse ways as well at this point
    if keep_ways is False and keep_relations is True:
        pass
    # If ways are not wanted, neither should relations be parsed
    elif not keep_ways:
        way_records = None
        relations = None

    if not keep_relations:
        relations = None

    # Call signature for fetching POIs
    nodes, ways, relation_ways, relations = get_osm_data(node_arrays=nodes,
                                                         way_records=way_records,
                                                         relations=relations,
                                                         tags_as_columns=tags_as_columns,
                                                         data_filter=custom_filter,
                                                         filter_type=filter_type,
                                                         osm_keys=osm_keys,
                                                         )

    # If there weren't any data, return empty GeoDataFrame
    if nodes is None and ways is None and relations is None:
        warnings.warn("Could not find any OSM data for given area.",
                      UserWarning,
                      stacklevel=2)
        return None

    # Ensure that ways are None if returning those are not requested
    if not keep_ways:
        ways = None

    # Prepare GeoDataFrame
    gdf = prepare_geodataframe(nodes, node_coordinates, ways,
                               relations, relation_ways,
                               tags_as_columns, bounding_box)

    return gdf



from pyrosm.data_manager import get_osm_data
from pyrosm.frames import prepare_geodataframe
from pyrosm.utils import validate_custom_filter
import warnings


def get_poi_data(nodes, node_coordinates, way_records, relations, tags_as_columns,
                 custom_filter, bounding_box):
    # Validate filter
    custom_filter = validate_custom_filter(custom_filter)

    # Call signature for fetching POIs
    nodes, ways, relation_ways, relations = get_osm_data(node_arrays=nodes,
                                                         way_records=way_records,
                                                         relations=relations,
                                                         tags_as_columns=tags_as_columns,
                                                         data_filter=custom_filter,
                                                         filter_type="keep",
                                                         osm_keys=None,
                                                         )

    # If there weren't any data, return empty GeoDataFrame
    if nodes is None and ways is None and relations is None:
        warnings.warn("Could not find any POIs for given area.",
                      UserWarning,
                      stacklevel=2)
        return None

    # Prepare GeoDataFrame
    gdf = prepare_geodataframe(nodes, node_coordinates, ways,
                               relations, relation_ways,
                               tags_as_columns, bounding_box)

    return gdf



from shapely.geos import geos_version_string as shapely_geos_version
from pygeos import geos_capi_version_string
import warnings

# shapely has something like: "3.6.2-CAPI-1.10.2 4d2925d6"
# pygeos has something like: "3.6.2-CAPI-1.10.2"
if not shapely_geos_version.startswith(geos_capi_version_string):
    warnings.warn(
        "The Shapely GEOS version ({}) is incompatible with the GEOS "
        "version PyGEOS was compiled with ({}). The tool will work "
        "but it runs a bit slower.".format(
            shapely_geos_version, geos_capi_version_string
        )
    )
    PYGEOS_SHAPELY_COMPAT = False
else:
    PYGEOS_SHAPELY_COMPAT = True



from shapely import ops
from shapely.geometry import MultiLineString, \
    Polygon, MultiPolygon, box
from pyrosm_proto import BlobHeader, Blob, HeaderBlock
from pyrosm.exceptions import PBFNotImplemented
import zlib
from struct import unpack
import os


def validate_custom_filter(custom_filter):
    # Check that the custom filter is in correct format
    if not isinstance(custom_filter, dict):
        raise ValueError(f"'custom_filter' should be a Python dictionary. "
                         f"Got {custom_filter} with type {type(custom_filter)}.")

    for k, v in custom_filter.items():
        if not isinstance(k, str):
            raise ValueError(f"OSM key in 'custom_filter' should be string. "
                             f"Got {k} of type {type(k)}")
        if v is True:
            custom_filter[k] = [v]
            continue

        if not isinstance(v, list):
            raise ValueError(f"OSM tags in 'custom_filter' should be inside a list. "
                             f"Got {v} of type {type(v)}")

        for item in v:
            if item is True:
                continue
            if not isinstance(item, str):
                raise ValueError(f"OSM tag (value) in 'custom_filter' should be string. "
                                 f"Got {item} of type {type(item)}")
    return custom_filter


def validate_osm_keys(osm_keys):
    if osm_keys is not None:
        if type(osm_keys) not in [str, list]:
            raise ValueError(f"'osm_keys_to_keep' -parameter should be of type str or list. "
                             f"Got {osm_keys} of type {type(osm_keys)}.")


def validate_tags_as_columns(tags_as_columns):
    if not isinstance(tags_as_columns, list):
        raise ValueError(f"'tags_as_columns' should be a list. "
                         f"Got {tags_as_columns} of type {type(tags_as_columns)}.")
    for col in tags_as_columns:
        if not isinstance(col, str):
            raise ValueError(f"All tags listed in 'tags_as_columns' should be strings. "
                             f"Got {col} of type {type(col)}.")


def validate_booleans(keep_nodes, keep_ways, keep_relations):
    if not isinstance(keep_nodes, bool):
        raise ValueError("'keep_nodes' should be boolean type: True or False")

    if not isinstance(keep_ways, bool):
        raise ValueError("'keep_ways' should be boolean type: True or False")

    if not isinstance(keep_relations, bool):
        raise ValueError("'keep_relations' should be boolean type: True or False")

    if keep_nodes is False and keep_ways is False and keep_relations is False:
        raise ValueError("At least on of the following parameters should be True: "
                         "'keep_nodes', 'keep_ways', or 'keep_relations'")


def validate_boundary_type(boundary_type):
    allowed_boundary_types = ["administrative", "national_park", "political",
                              "postal_code", "protected_area", "aboriginal_lands",
                              "maritime", "marker",
                              # There is no consensus whether allowing the following ones should be done
                              # but as they exist, allow using them here as well.
                              # https://wiki.openstreetmap.org/wiki/Parcel
                              "lot", "parcel", "tract",
                              "all"]
    allowed_text = ", ".join(allowed_boundary_types)
    if not isinstance(boundary_type, str):
        raise ValueError(f"'boundary_type' should be one of the following: {allowed_text}."
                         f"Got '{boundary_type}' of type {type(boundary_type)}.")

    boundary_type = boundary_type.strip().lower()
    if boundary_type not in allowed_boundary_types:
        raise ValueError(f"'boundary_type' should be one of the following: {allowed_text}."
                         f"Got '{boundary_type}' of type {type(boundary_type)}.")
    return boundary_type


def validate_bounding_box(geom):
    if type(geom) in [Polygon, MultiPolygon]:
        return geom

    elif isinstance(geom, MultiLineString):
        geom = ops.linemerge(geom)

    if not geom.is_closed:
        raise ValueError(
            "Provided bounding box is not a closed geometry. "
            "Ensure that you pass a Polygon or LinearRing."
        )
    return Polygon(geom)


def validate_input_file(filepath):
    if not isinstance(filepath, str):
        raise ValueError("'filepath' should be a string.")
    if not filepath.endswith(".pbf"):
        raise ValueError(f"Input data should be in Protobuf format (*.osm.pbf). "
                         f"Found: {filepath.split('.')[-1]}")
    if not os.path.exists(filepath):
        raise ValueError(f"File does not exist: "
                         f"Found: {filepath}")
    return filepath


def get_bounding_box(filepath):
    with open(filepath, 'rb') as f:

        # Check that the data stream is valid OSM
        # =======================================

        buf = f.read(4)
        msg_len = unpack('!L', buf)[0]
        msg = BlobHeader()
        msg.ParseFromString(f.read(msg_len))
        blob_header = msg

        msg = Blob()
        msg.ParseFromString(f.read(blob_header.datasize))
        blob_data = zlib.decompress(msg.zlib_data)
        header_block = HeaderBlock()
        header_block.ParseFromString(blob_data)

        for feature in header_block.required_features:
            if not (feature in ('OsmSchema-V0.6', 'DenseNodes')):
                raise PBFNotImplemented(
                    'Required feature %s not implemented!',
                    feature)

        # Parse bounding box
        try:
            bb = header_block.bbox.SerializeToDict()
            div = 1000000000
            bbox = box(bb["left"] / div,
                       bb["bottom"] / div,
                       bb["right"] / div,
                       bb["top"] / div,
                       )
        except Exception:
            bbox = None
        return bbox



from pyrosm.exceptions import PBFNotImplemented
from struct import unpack
import zlib
from pyrosm_proto import BlobHeader, Blob, HeaderBlock, PrimitiveBlock
from pyrosm.tagparser cimport tounicode, parse_dense_tags, parse_tags, explode_way_tags
from pyrosm._arrays cimport to_clong_array
from pyrosm.delta_compression cimport delta_decode_latitude, delta_decode_longitude, \
    delta_decode_id, delta_decode_timestamp, delta_decode_changeset
from pyrosm.data_filter cimport get_nodeid_lookup_khash, nodes_for_way_exist_khash
import numpy as np
from libc.stdlib cimport malloc, free

cdef get_primitive_blocks_and_string_tables(filepath):
    cdef int msg_len
    cdef bytes blob_data
    cdef str feature

    with open(filepath, 'rb') as f:

        # Check that the data stream is valid OSM
        # =======================================

        buf = f.read(4)
        msg_len = unpack('!L', buf)[0]
        msg = BlobHeader()
        msg.ParseFromString(f.read(msg_len))
        blob_header = msg

        msg = Blob()
        msg.ParseFromString(f.read(blob_header.datasize))
        blob_data = zlib.decompress(msg.zlib_data)
        header_block = HeaderBlock()
        header_block.ParseFromString(blob_data)

        for feature in header_block.required_features:
            if not (feature in ('OsmSchema-V0.6', 'DenseNodes')):
                raise PBFNotImplemented(
                    'Required feature %s not implemented!',
                    feature)

        # Gather primitive blocks and string tables
        primitive_blocks = []
        string_tables = []

        while True:
            # Read header
            buf = f.read(4)

            # Stop when the end has been reached
            if len(buf) == 0:
                break

            msg_len = unpack('!L', buf)[0]

            msg = BlobHeader()
            msg.ParseFromString(f.read(msg_len))
            blob_header = msg

            # Get data
            msg = Blob()
            msg.ParseFromString(f.read(blob_header.datasize))
            blob_data = zlib.decompress(msg.zlib_data)

            # Get primite block
            pblock = PrimitiveBlock()
            pblock.ParseFromString(blob_data)

            # Get string table and decode
            str_table = [tounicode(s) for s in pblock.stringtable.s]

            primitive_blocks.append(pblock)
            string_tables.append(str_table)

    return primitive_blocks, string_tables


cdef parse_dense(pblock, data, string_table, bounding_box):
    cdef:
        int node_granularity = pblock.granularity
        int timestamp_granularity = pblock.date_granularity
        int lon_offset = pblock.lon_offset
        int lat_offset = pblock.lat_offset

    # Get latitudes
    lats = delta_decode_latitude(data, node_granularity, lat_offset)

    # Get longitudes
    lons = delta_decode_longitude(data, node_granularity, lon_offset)

    # Version
    versions = np.array(list(data.denseinfo.version), dtype=np.int64)

    # Ids
    ids = delta_decode_id(data)

    # Timestamp
    timestamps = delta_decode_timestamp(data, timestamp_granularity)

    # Changeset
    changesets = delta_decode_changeset(data)

    # Tags
    tags = np.empty(len(data.id), dtype=object)
    parsed = parse_dense_tags(data.keys_vals, string_table)

    # In some cases node-tags are not available at all
    if len(parsed) != 0:
        tags[:] = parsed

    # Metadata might not be available, if so add empty
    # This can happen with BBBike data
    if versions.shape[0] != 0:
        versions = np.empty(len(data.id), dtype=np.int8)
    if changesets.shape[0] != 0:
        changesets = np.empty(len(data.id), dtype=np.int8)
    if timestamps.shape[0] != 0:
        timestamps = np.empty(len(data.id), dtype=np.int8)

    if bounding_box is not None:
        # Filter
        xmin, ymin, xmax, ymax = bounding_box
        mask = (xmin <= lons) & (lons <= xmax) & (ymin <= lats) & (lats <= ymax)
        ids = ids[mask]
        versions = versions[mask]
        changesets = changesets[mask]
        timestamps = timestamps[mask]
        lons = lons[mask]
        lats = lats[mask]
        tags = tags[mask]

    return [dict(id=ids,
                 version=versions,
                 changeset=changesets,
                 timestamp=timestamps,
                 lon=lons,
                 lat=lats,
                 tags=tags,
                 )]


cdef parse_nodes(pblock, data, bounding_box):
    ids = []
    versions = []
    changesets = []
    timestamps = []
    lons = []
    lats = []

    granularity = pblock.granularity
    lon_offset = pblock.lon_offset
    lat_offset = pblock.lat_offset
    div = 1000000000

    for node in data:
        try:
            changesets.append(int(node.info.changeset))
        except:
            changesets.append(0)
        versions.append(node.info.version)
        ids.append(node.id)
        timestamps.append(node.info.timestamp)
        lons.append(node.lon)
        lats.append(node.lat)

    id_ = np.array(ids, dtype=np.int64)
    version = np.array(versions, dtype=np.int64)
    changeset = np.array(changesets, dtype=np.int64)
    timestamp = np.array(timestamps, dtype=np.int64)
    lon = (np.array(lons, dtype=np.int64) * granularity + lon_offset) / div
    lat = (np.array(lats, dtype=np.int64) * granularity + lat_offset) / div

    if bounding_box is not None:
        # Filter
        xmin, ymin, xmax, ymax = bounding_box
        mask = (xmin <= lon) & (lon <= xmax) & (ymin <= lat) & (lat <= ymax)
        id_ = id_[mask]
        version = version[mask]
        changeset = changeset[mask]
        timestamp = timestamp[mask]
        lon = lon[mask]
        lat = lat[mask]

    return dict(id=id_,
                version=version,
                changeset=changeset,
                timestamp=timestamp,
                lon=lon,
                lat=lat
                )


cdef parse_nodeids_from_ref_deltas(refs):
    cdef long long delta, nid = 0
    cdef int i, N = len(refs)
    cdef long long * refs_c = to_clong_array(refs)
    cdef long long *nodes = <long long *>malloc(N*sizeof(long long))

    try:
        for i in range(N):
            nid += refs_c[i]
            nodes[i] = nid
        return [x for x in nodes[:N]]

    finally:
        free(nodes)
        free(refs_c)


cdef parse_ways(data, string_table, node_lookup):
    cdef long long id
    cdef int version, i, timestamp, n=len(data)

    way_set = []
    for i in range(0, n):
        way = data[i]
        nodes = parse_nodeids_from_ref_deltas(way.refs)
        if node_lookup is not None:
            if nodes_for_way_exist_khash(nodes, node_lookup):
                way_set.append(
                    dict(
                        id=way.id,
                        version=way.info.version,
                        timestamp=way.info.timestamp,
                        tags=parse_tags(way.keys, way.vals, string_table),
                        nodes=nodes,
                    )
                )
        else:
            way_set.append(
                dict(
                    id=way.id,
                    version=way.info.version,
                    timestamp=way.info.timestamp,
                    tags=parse_tags(way.keys, way.vals, string_table),
                    nodes=nodes,
                )
            )
    return way_set


cdef get_relation_members(relation, str_table):
    relation_types = {
        0: b"node",
        1: b"way",
        2: b"relation"
        }

    # Member-ids
    member_N = len(relation.memids)
    member_id_deltas = np.zeros(member_N+1, dtype=np.int64)
    member_id_deltas[1:] = list(relation.memids)
    member_ids = np.cumsum(member_id_deltas)[1:]

    # Member types - 0: Node, 1: Way, 2: Relation
    member_types = np.array([relation_types[t]
                             for t in relation.types],
                            dtype=object)

    # Member roles
    member_roles = np.array([str_table[role_id] for role_id in relation.roles_sid],
                            dtype=object)

    return dict(
        member_id=member_ids,
        member_type=member_types,
        member_role=member_roles,
        )


cdef parse_relations(data, string_table):
    N = len(data)

    # Version
    versions = np.array([rel.info.version for rel in data], dtype=np.int64)

    # Ids (delta coded)
    id_deltas = np.zeros(N+1, dtype=np.int64)
    id_deltas[1:] = [rel.id for rel in data]
    ids = np.cumsum(id_deltas)[1:]

    # Timestamp
    timestamps = np.array([rel.info.timestamp for rel in data], dtype=np.int64)

    # Changeset
    changesets = np.array([rel.info.changeset for rel in data], dtype=np.int64)

    # Relation members
    members = np.array([
        get_relation_members(rel, string_table)
        for rel in data],
        dtype=object
        )

    # Tags
    tags = np.array([
            parse_tags(rel.keys, rel.vals, string_table)
            for rel in data],
            dtype=object
            )

    return [dict(id=ids,
                version=versions,
                changeset=changesets,
                timestamp=timestamps,
                members=members,
                tags=tags,
                )]


cpdef parse_osm_data(filepath, bounding_box,
                     exclude_relations):
    return _parse_osm_data(filepath, bounding_box, exclude_relations)


cdef _parse_osm_data(filepath, bounding_box,
                     exclude_relations):
    all_ways = []
    all_nodes = []
    all_relations = []
    node_lookup_created = False

    primitive_blocks, string_tables = get_primitive_blocks_and_string_tables(filepath)

    for pblock, str_table in zip(primitive_blocks, string_tables):
        for pgroup in pblock.primitivegroup:
            if len(pgroup.dense.id) > 0:
                all_nodes += parse_dense(pblock, pgroup.dense, str_table, bounding_box)
            elif len(pgroup.nodes) > 0:
                all_nodes += parse_nodes(pblock, pgroup.nodes, bounding_box)
            elif len(pgroup.ways) > 0:
                # Once all the nodes have been parsed comes Ways
                if bounding_box is not None:
                    if not node_lookup_created:
                        node_lookup = get_nodeid_lookup_khash(all_nodes)
                        all_ways += parse_ways(pgroup.ways, str_table, node_lookup)
                    else:
                        all_ways += parse_ways(pgroup.ways, str_table, node_lookup)
                else:
                    all_ways += parse_ways(pgroup.ways, str_table, None)
            elif len(pgroup.relations) > 0:
                if exclude_relations:
                    continue
                all_relations += parse_relations(pgroup.relations, str_table)

    # Explode the way tags
    all_ways, all_way_tag_keys = explode_way_tags(all_ways)

    return all_nodes, all_ways, all_relations, all_way_tag_keys



# BASE
URL = "https://download.bbbike.org/osm/bbbike"
suffix = ".osm.pbf"


class Cities:
    available = ['Aachen',
                 'Aarhus',
                 'Adelaide',
                 'Albuquerque',
                 'Alexandria',
                 'Amsterdam',
                 'Antwerpen',
                 'Arnhem',
                 'Auckland',
                 'Augsburg',
                 'Austin',
                 'Baghdad',
                 'Baku',
                 'Balaton',
                 'Bamberg',
                 'Bangkok',
                 'Barcelona',
                 'Basel',
                 'Beijing',
                 'Beirut',
                 'Berkeley',
                 'Berlin',
                 'Bern',
                 'Bielefeld',
                 'Birmingham',
                 'Bochum',
                 'Bogota',
                 'Bombay',
                 'Bonn',
                 'Bordeaux',
                 'Boulder',
                 'BrandenburgHavel',
                 'Braunschweig',
                 'Bremen',
                 'Bremerhaven',
                 'Brisbane',
                 'Bristol',
                 'Brno',
                 'Bruegge',
                 'Bruessel',
                 'Budapest',
                 'BuenosAires',
                 'Cairo',
                 'Calgary',
                 'Cambridge',
                 'CambridgeMa',
                 'Canberra',
                 'CapeTown',
                 'Chemnitz',
                 'Chicago',
                 'ClermontFerrand',
                 'Colmar',
                 'Copenhagen',
                 'Cork',
                 'Corsica',
                 'Corvallis',
                 'Cottbus',
                 'Cracow',
                 'CraterLake',
                 'Curitiba',
                 'Cusco',
                 'Dallas',
                 'Darmstadt',
                 'Davis',
                 'DenHaag',
                 'Denver',
                 'Dessau',
                 'Dortmund',
                 'Dresden',
                 'Dublin',
                 'Duesseldorf',
                 'Duisburg',
                 'Edinburgh',
                 'Eindhoven',
                 'Emden',
                 'Erfurt',
                 'Erlangen',
                 'Eugene',
                 'Flensburg',
                 'FortCollins',
                 'Frankfurt',
                 'FrankfurtOder',
                 'Freiburg',
                 'Gdansk',
                 'Genf',
                 'Gent',
                 'Gera',
                 'Glasgow',
                 'Gliwice',
                 'Goerlitz',
                 'Goeteborg',
                 'Goettingen',
                 'Graz',
                 'Groningen',
                 'Halifax',
                 'Halle',
                 'Hamburg',
                 'Hamm',
                 'Hannover',
                 'Heilbronn',
                 'Helsinki',
                 'Hertogenbosch',
                 'Huntsville',
                 'Innsbruck',
                 'Istanbul',
                 'Jena',
                 'Jerusalem',
                 'Johannesburg',
                 'Kaiserslautern',
                 'Karlsruhe',
                 'Kassel',
                 'Katowice',
                 'Kaunas',
                 'Kiel',
                 'Kiew',
                 'Koblenz',
                 'Koeln',
                 'Konstanz',
                 'LaPaz',
                 'LaPlata',
                 'LakeGarda',
                 'Lausanne',
                 'Leeds',
                 'Leipzig',
                 'Lima',
                 'Linz',
                 'Lisbon',
                 'Liverpool',
                 'Ljubljana',
                 'Lodz',
                 'London',
                 'Luebeck',
                 'Luxemburg',
                 'Lyon',
                 'Maastricht',
                 'Madison',
                 'Madrid',
                 'Magdeburg',
                 'Mainz',
                 'Malmoe',
                 'Manchester',
                 'Mannheim',
                 'Marseille',
                 'Melbourne',
                 'Memphis',
                 'MexicoCity',
                 'Miami',
                 'Moenchengladbach',
                 'Montevideo',
                 'Montpellier',
                 'Montreal',
                 'Moscow',
                 'Muenchen',
                 'Muenster',
                 'NewDelhi',
                 'NewOrleans',
                 'NewYorkCity',
                 'Nuernberg',
                 'Oldenburg',
                 'Oranienburg',
                 'Orlando',
                 'Oslo',
                 'Osnabrueck',
                 'Ostrava',
                 'Ottawa',
                 'Paderborn',
                 'Palma',
                 'PaloAlto',
                 'Paris',
                 'Perth',
                 'Philadelphia',
                 'PhnomPenh',
                 'Portland',
                 'PortlandME',
                 'Porto',
                 'PortoAlegre',
                 'Potsdam',
                 'Poznan',
                 'Prag',
                 'Providence',
                 'Regensburg',
                 'Riga',
                 'RiodeJaneiro',
                 'Rostock',
                 'Rotterdam',
                 'Ruegen',
                 'Saarbruecken',
                 'Sacramento',
                 'Saigon',
                 'Salzburg',
                 'SanFrancisco',
                 'SanJose',
                 'SanktPetersburg',
                 'SantaBarbara',
                 'SantaCruz',
                 'Santiago',
                 'Sarajewo',
                 'Schwerin',
                 'Seattle',
                 'Seoul',
                 'Sheffield',
                 'Singapore',
                 'Sofia',
                 'Stockholm',
                 'Stockton',
                 'Strassburg',
                 'Stuttgart',
                 'Sucre',
                 'Sydney',
                 'Szczecin',
                 'Tallinn',
                 'Tehran',
                 'Tilburg',
                 'Tokyo',
                 'Toronto',
                 'Toulouse',
                 'Trondheim',
                 'Tucson',
                 'Turin',
                 'UlanBator',
                 'Ulm',
                 'Usedom',
                 'Utrecht',
                 'Vancouver',
                 'Victoria',
                 'WarenMueritz',
                 'Warsaw',
                 'WashingtonDC',
                 'Waterloo',
                 'Wien',
                 'Wroclaw',
                 'Wuerzburg',
                 'Wuppertal',
                 'Zagreb',
                 'Zuerich']

    available.sort()

    # Create data sources
    _sources = {
        city.lower(): {"name": city + suffix, "url": f"{URL}/{city}/{city}{suffix}"}
        for city in available if city not in ["NewYorkCity"]
    }
    # Add New York City separately as there is also a state with the same name
    # NewYork ==> NewYorkCity
    _ny = "newyorkcity"
    _ny_src = "NewYork"
    _sources[_ny] = {"name": _ny + suffix, "url": f"{URL}/{_ny_src}/{_ny_src}{suffix}"}

    __dict__ = _sources

    def __getattr__(self, name):
        return self.__dict__[name.lower()]



# ===============================
# OSM TAGS
# ===============================
# These configurations are used to specify which OSM tags are kept
# by default as columns in DataFrame. Follows more or less the
# OSM documentation and the tags mentioned on the Wiki:
# https://wiki.openstreetmap.org/wiki/Map_Features

# If a tag is NOT mentioned in these lists (below),
# such tag will be kept in a JSON that is stored in a "tag" column.
# The purpose of specifying default tag structure is to speed up the parsing, and
# reduce the complexity in the resulting DataFrame as OSM data can contain numerous
# attributes that are not most relevant for basic user (e.g. "customized" tags).

# ========================
# BASIC INFO TAGS
# ========================
# A list of tags that are relevant for most OSM keys
basic_info_tags = ["addr:city",
                   "addr:country",
                   "addr:full",
                   "addr:housenumber",
                   "addr:housename",
                   "addr:postcode",
                   "addr:place",
                   "addr:street",
                   "email",
                   "name",
                   "opening_hours",
                   "operator",
                   "phone",
                   "ref",
                   "url",
                   "website",
                   "yes",
                   ]

# ========================
# HIGHWAY TAGS
# ========================

# Default tags to keep as columns with highways
# Mostly based on: https://wiki.openstreetmap.org/wiki/Key:highway
highway_columns = ["access",
                   "area",
                   "bicycle",
                   "bicycle_road",
                   "bridge",
                   "busway",
                   "cycleway",
                   "est_width",
                   "foot",
                   "footway",
                   "highway",
                   "int_ref",
                   "junction",
                   "lanes",
                   "lit",
                   "maxspeed",
                   "motorcar",
                   "motorroad",
                   "motor_vehicle",
                   "name",
                   "oneway",
                   "overtaking",
                   "path",
                   "passing_places",
                   "psv",
                   "ref",
                   "service",
                   "segregated",
                   "sidewalk",
                   "smoothness",
                   "surface",
                   "tracktype",
                   "tunnel",
                   "turn",
                   "width",
                   "winter_road",

                   # Other highway tags which are not kept by default
                   # (more tags slows down the parsing)
                   # =====================================

                   # "abutters",
                   # "driving_side",
                   # "embedded_rails",
                   # "ford",
                   # "ice_road",
                   # "incline",
                   # "mtb:scale",
                   # "mtb:scale:uphill",
                   # "mtb:scale:imba",
                   # "mtb:description",
                   # "parking:condition",
                   # "parking:lane",
                   # "sac_scale",
                   # "tactile_paving",
                   # "traffic_calming",
                   # "trail_visibility",

                   ]

# ========================
# BUILDING TAGS
# ========================
# See:
# https://wiki.openstreetmap.org/wiki/Key:building
# https://wiki.openstreetmap.org/wiki/Key:addr

# Default tags to keep as columns with buildings

building_columns = basic_info_tags + \
                   ["building",
                    "amenity",
                    "building:flats",
                    "building:levels",
                    "building:material",
                    "building:max_level",
                    "building:min_level",
                    "building:fireproof",
                    "building:use",
                    "craft",
                    "height",
                    "internet_access",
                    "landuse",
                    "levels",
                    "office",
                    "operator",
                    "shop",
                    "source",
                    "start_date",
                    "wikipedia",

                    # Other tags that are not kept by default
                    # "addr:conscriptionnumber",
                    # "addr:district",
                    # "addr:province",
                    # "soft_storey",
                    # "takeaway",
                    # "addr:state",
                    # "mml:class"
                    ]

# ========================
# AMENITY TAGS
# ========================
# See:
# https://wiki.openstreetmap.org/wiki/Key:amenity

# Default tags to keep as columns with amenities
amenity_columns = basic_info_tags + \
                  ["amenity",
                   "arts_centre",
                   "atm",
                   "bank",
                   "bicycle_parking",
                   "bicycle_rental",
                   "bicycle_repair_station",
                   "bar",
                   "bbq",
                   "biergarten",
                   "brothel",
                   "building",
                   "building:levels",
                   "bureau_de_change",
                   "bus_station",
                   "bus_stop",
                   "cafe",
                   "car_rental",
                   "car_repair",
                   "car_sharing",
                   "car_wash",
                   "casino",
                   "charging_station",
                   "childcare",
                   "cinema",
                   "clinic",
                   "college",
                   "dentist",
                   "doctors",
                   "driving_school",
                   "drinking_water",
                   "fast_food",
                   "ferry_terminal",
                   "fire_station",
                   "food_court",
                   "fountain",
                   "fuel",
                   "gambling",
                   "hospital",
                   "ice_cream",
                   "internet_access",
                   "kindergarten",
                   "landuse",
                   "language_school",
                   "library",
                   "music_school",
                   "nightclub",
                   "nursing_home",
                   "office",
                   "operator",
                   "parking",
                   "pharmasy",
                   "planetarium",
                   "police",
                   "post_office",
                   "pub",
                   "public_bath",
                   "rescue_station",
                   "restaurant",
                   "retirement_home",
                   "school",
                   "social_centre",
                   "social_facility",
                   "source",
                   "spa",
                   "start_date",
                   "stripclub",
                   "taxi",
                   "theatre",
                   "university",
                   "wikipedia",

                   # Other tags
                   # ----------
                   # "baby_hatch",
                   # "levels",
                   # "ev_charging",
                   # "motorcycle_parking",
                   # "parking_entrance",
                   # "parking_space",
                   # "dive_centre",
                   # "dojo",
                   ]

# ========================
# SHOP TAGS
# ========================
# See:
# https://wiki.openstreetmap.org/wiki/Key:shop

# Default tags to keep as columns with amenities
shop_columns = basic_info_tags + \
               ["agrarian",
                "alcohol",
                "anime",
                "antiques",
                "appliance",
                "art",
                "atv",
                "baby_goods",
                "bag",
                "bakery",
                "bathroom_furnishing",
                "bed",
                "beverages",
                "bicycle",
                "boat",
                "bookmaker",
                "books",
                "boutique",
                "brewing_supplies",
                "butcher",
                "camera",
                "candles",
                "cannabis",
                "car",
                "car_parts",
                "car_repair",
                "caravan",
                "carpet",
                "charity",
                "cheese",
                "chemist",
                "chocolate",
                "clothes",
                "coffee",
                "collector",
                "computer",
                "confectionery",
                "convenience",
                "copyshop",
                "cosmetics",
                "craft",
                "curtain",
                "dairy",
                "deli",
                "department_store",
                "doityourself",
                "doors",
                "drugstore",
                "dry_cleaning",
                "e-cigarette",
                "electrical",
                "electronics",
                "energy",
                "erotic",
                "fabric",
                "farm",
                "fashion",
                "fireplace",
                "fishing",
                "flooring",
                "florist",
                "frame",
                "free_flying",
                "frozen_food",
                "fuel",
                "funeral_directors",
                "furniture",
                "games",
                "garden_centre",
                "garden_furniture",
                "gas",
                "general",
                "gift",
                "glaziery",
                "golf",
                "greengrocer",
                "hairdresser",
                "hairdresser_supply",
                "hardware",
                "health_food",
                "hearing_aids",
                "herbalist",
                "hifi",
                "houseware",
                "hunting",
                "ice_cream",
                "interior_decoration",
                "jetski",
                "jewelry",
                "kiosk",
                "kitchen",
                "lamps",
                "laundry",
                "leather",
                "lighting",
                "locksmith",
                "lottery",
                "mall",
                "massage",
                "medical_supply",
                "military_surplus",
                "mobile_phone",
                "model",
                "money_lender",
                "motorcycle",
                "music",
                "musical_instrument",
                "name",
                "newsagent",
                "nutrition_supplements",
                "optician",
                "organic",
                "outdoor",
                "outpost",
                "paint",
                "party",
                "pasta",
                "pastry",
                "pawnbroker",
                "perfumery",
                "pest_control",
                "pet",
                "pet_grooming",
                "phone",
                "photo",
                "pyrotechnics",
                "radiotechnics",
                "religion",
                "robot",
                "scuba_diving",
                "seafood",
                "second_hand",
                "security",
                "sewing",
                "shoes",
                "shop",
                "ski",
                "snowmobile",
                "spices",
                "sports",
                "stationery",
                "storage_rental",
                "supermarket",
                "swimming_pool",
                "tailor",
                "tattoo",
                "tea",
                "ticket",
                "tiles",
                "tobacco",
                "toys",
                "trade",
                "trailer",
                "travel_agency",
                "trophy",
                "tyres",
                "user defined",
                "vacant",
                "vacuum_cleaner",
                "variety_store",
                "video",
                "video_games",
                "watches",
                "water",
                "weapons",
                "wholesale",
                "window_blind",
                "windows",
                "wine",
                "wool",
                ]

# ========================
# CRAFT TAGS
# ========================
# See: https://wiki.openstreetmap.org/wiki/Key%3Acraft
craft_columns = basic_info_tags + \
                ["agricultural_engines",
                 "atelier",
                 "bakery",
                 "basket_maker",
                 "beekeeper",
                 "blacksmith",
                 "boatbuilder",
                 "bookbinder",
                 "brewery",
                 "builder",
                 "cabinet_maker",
                 "carpenter",
                 "carpet_layer",
                 "caterer",
                 "chimney_sweeper",
                 "clockmaker",
                 "confectionery",
                 "cooper",
                 "craft",
                 "dental_technician",
                 "distillery",
                 "door_construction",
                 "dressmaker",
                 "electronics_repair",
                 "embroiderer",
                 "electrician",
                 "engraver",
                 "floorer",
                 "gardener",
                 "glaziery",
                 "grinding_mill",
                 "handicraft",
                 "hvac",
                 "insulation",
                 "jeweller",
                 "joiner",
                 "key_cutter",
                 "locksmith",
                 "metal_construction",
                 "mint",
                 "musical_instrument",
                 "oil_mill",
                 "optician",
                 "organ_builder",
                 "painter",
                 "parquet_layer",
                 "photographer",
                 "photographic_laboratory",
                 "piano_tuner",
                 "plasterer",
                 "plumber",
                 "pottery",
                 "printer",
                 "printmaker",
                 "rigger",
                 "roofer",
                 "saddler",
                 "sailmaker",
                 "sawmill",
                 "scaffolder",
                 "sculptor",
                 "shoemaker",
                 "signmaker",
                 "stand_builder",
                 "stonemason",
                 "sun_protection",
                 "tailor",
                 "tiler",
                 "tinsmith",
                 "toolmaker",
                 "turner",
                 "upholsterer",
                 "watchmaker",
                 "water_well_drilling",
                 "window_construction",
                 "winery",
                 ]

# ========================
# LEISURE TAGS
# ========================
# See: https://wiki.openstreetmap.org/wiki/Key%3Acraft
leisure_columns = basic_info_tags + \
                  ["adult_gaming_centre",
                   "amusement_arcade",
                   "bandstand",
                   "beach_resort",
                   "bird_hide",
                   "bowling_alley",
                   "common",
                   "dance",
                   "disc_golf_course",
                   "dog_park",
                   "escape_game",
                   "firepit",
                   "fishing",
                   "fitness_centre",
                   "fitness_station",
                   "garden",
                   "golf_course",
                   "hackerspace",
                   "horse_riding",
                   "ice_rink",
                   "leisure",
                   "marina",
                   "miniature_golf",
                   "nature_reserve",
                   "outdoor_seating",
                   "park",
                   "picnic_table",
                   "pitch",
                   "playground",
                   "sauna",
                   "slipway",
                   "sports_centre",
                   "stadium",
                   "summer_camp",
                   "swimming_area",
                   "swimming_pool",
                   "track",
                   "trampoline_park",
                   "water_park",
                   "wildlife_hide",
                   ]

# ========================
# TOURISM TAGS
# ========================
# See: https://wiki.openstreetmap.org/wiki/Key%3Atourism
tourism_columns = basic_info_tags + \
                  ["alpine_hut",
                   "apartment",
                   "aquarium",
                   "artwork",
                   "attraction",
                   "camp_pitch",
                   "camp_site",
                   "caravan_site",
                   "chalet",
                   "gallery",
                   "guest_house",
                   "hostel",
                   "hotel",
                   "information",
                   "motel",
                   "museum",
                   "picnic_site",
                   "theme_park",
                   "tourism",
                   "viewpoint",
                   "wilderness_hut",
                   "zoo",
                   ]

# ========================
# HISTORIC TAGS
# ========================
# See: https://wiki.openstreetmap.org/wiki/Key%3Ahistoric
historic_columns = basic_info_tags + \
                   ["aircraft",
                    "aqueduct",
                    "archaeological_site",
                    "battlefield",
                    "boundary_stone",
                    "building",
                    "cannon",
                    "castle",
                    "castle_wall",
                    "church",
                    "city_gate",
                    "citywalls",
                    "farm",
                    "fort",
                    "gallows",
                    "highwater_mark",
                    "historic",
                    "locomotive",
                    "manor",
                    "memorial",
                    "milestone",
                    "monastery",
                    "monument",
                    "optical_telegraph",
                    "pillory",
                    "railway_car",
                    "ruins",
                    "rune_stone",
                    "ship",
                    "tank",
                    "tomb",
                    "tower",
                    "wayside_cross",
                    "wayside_shrine",
                    "wreck",
                    ]

# ========================
# EMERGENCY TAGS
# ========================
# See: https://wiki.openstreetmap.org/wiki/Key%3Aemergency
emergency_columns = basic_info_tags + \
                    ["access_point",
                     "ambulance_station",
                     "assembly_point",
                     "defibrillator",
                     "drinking_water",
                     "dry_riser_inlet",
                     "emergency",
                     "emergency_ward_entrance",
                     "fire_alarm_box",
                     "fire_extinguisher",
                     "fire_flapper",
                     "fire_hose",
                     "fire_hydrant",
                     "fire_water_pond",
                     "first_aid_kit",
                     "landing_site",
                     "life_ring",
                     "lifeguard",
                     "lifeguard_base",
                     "lifeguard_platform",
                     "lifeguard_tower",
                     "mountain_rescue",
                     "phone",
                     "rescue_box",
                     "siren",
                     "suction_point",
                     "water_tank",
                     ]

# ========================
# OFFICE TAGS
# ========================
# See: https://wiki.openstreetmap.org/wiki/Key%3Aoffice

office_columns = basic_info_tags + \
                 ["accountant",
                  "adoption_agency",
                  "advertising_agency",
                  "architect",
                  "association",
                  "bail_bond_agent",
                  "charity",
                  "company",
                  "consulting",
                  "coworking",
                  "diplomatic",
                  "educational_institution",
                  "employment_agency",
                  "energy_supplier",
                  "engineer",
                  "estate_agent",
                  "financial",
                  "forestry",
                  "foundation",
                  "geodesist",
                  "government",
                  "graphic_design",
                  "guide",
                  "harbour_master",
                  "insurance",
                  "it",
                  "lawyer",
                  "logistics",
                  "moving_company",
                  "newspaper",
                  "ngo",
                  "notary",
                  "office",
                  "political_party",
                  "private_investigator",
                  "property_management",
                  "quango",
                  "religion",
                  "research",
                  "security",
                  "surveyor",
                  "tax",
                  "tax_advisor",
                  "telecommunication",
                  "union",
                  "visa",
                  "water_utility",
                  ]

# ========================
# LANDUSE TAGS
# ========================
# See: https://wiki.openstreetmap.org/wiki/Key:landuse
landuse_columns = ["allotments",
                   "basin",
                   "brownfield",
                   "cemetery",
                   "commercial",
                   "conservation",
                   "construction",
                   "depot",
                   "farmland",
                   "farmyard",
                   "forest",
                   "garages",
                   "grass",
                   "greenfield",
                   "greenhouse_horticulture",
                   "industrial",
                   "landfill",
                   "landuse",
                   "meadow",
                   "military",
                   "orchard",
                   "peat_cutting",
                   "plant_nursery",
                   "port",
                   "quarry",
                   "railway",
                   "recreation_ground",
                   "religious",
                   "reservoir",
                   "residential",
                   "retail",
                   "salt_pond",
                   "village_green",
                   "vineyard",
                   "yes"]

# ========================
# NATURAL TAGS
# ========================
# See: https://wiki.openstreetmap.org/wiki/Key%3Anatural
natural_columns = ["arete",
                   "bare_rock",
                   "bay",
                   "beach",
                   "cape",
                   "cave_entrance",
                   "cliff",
                   "coastline",
                   "dune",
                   "fell",
                   "geyser",
                   "glacier",
                   "grassland",
                   "heath",
                   "hill",
                   "hot_spring",
                   "isthmus",
                   "moor",
                   "mud",
                   "natural",
                   "peak",
                   "peninsula",
                   "reef",
                   "ridge",
                   "rock",
                   "saddle",
                   "sand",
                   "scree",
                   "scrub",
                   "shingle",
                   "sinkhole",
                   "spring",
                   "stone",
                   "strait",
                   "tree",
                   "tree_row",
                   "valley",
                   "volcano",
                   "water",
                   "wetland",
                   "wood"
                   ]

# ========================
# WATERWAY TAGS
# ========================
# See: https://wiki.openstreetmap.org/wiki/Key%3Awaterway
waterway_columns = ["boatyard",
                    "canal",
                    "dam",
                    "ditch",
                    "dock",
                    "drain",
                    "drystream",
                    "fairway",
                    "fuel",
                    "lock_gate",
                    "pressurised",
                    "river",
                    "riverbank",
                    "stream",
                    "tidal_channel",
                    "turning_point",
                    "wadi",
                    "water_point",
                    "waterfall",
                    "waterway",
                    "weir",
                    ]

# ========================
# GEOLOGICAL TAGS
# ========================
# See: https://wiki.openstreetmap.org/wiki/Key%3Ageological
geological_columns = ["geological",
                      "moraine",
                      "outcrop",
                      "palaeontological_site",
                      ]

# ========================
# POWER TAGS
# ========================
# See: https://wiki.openstreetmap.org/wiki/Key%3Apower
power_columns = ["cable",
                 "catenary_mast",
                 "compensator",
                 "converter",
                 "generator",
                 "heliostat",
                 "insulator",
                 "line",
                 "busbar",
                 "bay",
                 "minor_line",
                 "plant",
                 "pole",
                 "portal",
                 "power",
                 "substation",
                 "switch",
                 "switchgear",
                 "terminal",
                 "tower",
                 "transformer",
                 ]

# ========================
# PUBLIC_TRANSPORT TAGS
# ========================
# See: https://wiki.openstreetmap.org/wiki/Key%3Apublic_transport
public_transport_columns = basic_info_tags + \
                           ["stop_position",
                            "platform",
                            "public_transport",
                            "station",
                            "stop_area",
                            ]

# ========================
# RAILWAY TAGS
# ========================
# See: https://wiki.openstreetmap.org/wiki/Key:railway
railway_columns = ["abandoned",
                   "buffer_stop",
                   "construction",
                   "crossing",
                   "derail",
                   "disused",
                   "funicular",
                   "halt",
                   "level_crossing",
                   "light_rail",
                   "miniature",
                   "monorail",
                   "narrow_gauge",
                   "platform",
                   "preserved",
                   "rail",
                   "railway",
                   "railway_crossing",
                   "roundhouse",
                   "signal",
                   "station",
                   "subway",
                   "subway_entrance",
                   "switch",
                   "tram",
                   "tram_stop",
                   "traverser",
                   "turntable",
                   "wash",
                   ]

# ========================
# ROUTE TAGS
# ========================
# See: https://wiki.openstreetmap.org/wiki/Key:route
route_columns = ["bicycle",
                 "bus",
                 "canoe",
                 "detour",
                 "duration",
                 "evacuation",
                 "ferry",
                 "foot",
                 "from",
                 "hiking",
                 "horse",
                 "inline_skates",
                 "light_rail",
                 "mtb",
                 "network",
                 "piste",
                 "railway",
                 "road",
                 "route",
                 "running",
                 "ski",
                 "subway",
                 "to",
                 "train",
                 "tracks",
                 "tram",
                 "trolleybus",
                 "type",
                 ]

# ========================
# PLACE TAGS
# ========================
# See: https://wiki.openstreetmap.org/wiki/Key%3Aplace
place_columns = ["allotments",
                 "archipelago",
                 "borough",
                 "city",
                 "city_block",
                 "continent",
                 "country",
                 "county",
                 "district",
                 "farm",
                 "hamlet",
                 "island",
                 "islet",
                 "isolated_dwelling",
                 "locality",
                 "municipality",
                 "neighbourhood",
                 "ocean",
                 "place",
                 "plot",
                 "province",
                 "quarter",
                 "region",
                 "sea",
                 "square",
                 "state",
                 "suburb",
                 "town",
                 "village"]

# ========================
# AERIALWAY TAGS
# ========================
# See: https://wiki.openstreetmap.org/wiki/Key%3Aaerialway
aerialway_columns = ["aerialway",
                     "cable_car",
                     "gondola",
                     "mixed_lift",
                     "chair_lift",
                     "drag_lift",
                     "t-bar",
                     "j-bar",
                     "platter",
                     "rope_tow",
                     "magic_carpet",
                     "zip_line",
                     "goods",
                     "pylon",
                     "station", ]

# ========================
# AEROWAY TAGS
# ========================
# See: https://wiki.openstreetmap.org/wiki/Key%3Aaeroway
aeroway_columns = ["aerodrome",
                   "aeroway",
                   "apron",
                   "control_tower",
                   "control_center",
                   "gate",
                   "hangar",
                   "helipad",
                   "heliport",
                   "navigationaid",
                   "beacon",
                   "runway",
                   "taxilane",
                   "taxiway",
                   "terminal",
                   "windsock",
                   "highway_strip",
                   ]


# ========================
# BOUNDARY TAGS
# ========================
# See: https://wiki.openstreetmap.org/wiki/Key:boundary
boundary_columns = basic_info_tags + \
                   ["aboriginal_lands",
                    "admin_level",
                    "administrative",
                    "boundary",
                    "left:district",
                    "right:district",
                    "maritime",
                    "marker",
                    "national_park",
                    "political",
                    "postal_code",
                    "protected_area",
                    "admin_level",
                    "border_type",
                    "start_date",
                   ]




def get_osm_filter(network_type):
    """
    Get OSM filter for different kinds of network types: 'driving', 'walking', 'cycling'.

    Applies same filters as in OSMnx,
    see: https://github.com/gboeing/osmnx/blob/master/osmnx/downloader.py#L19

    """
    if network_type == 'driving':
        return driving_filter()
    elif network_type == 'walking':
        return walking_filter()
    elif network_type == 'cycling':
        return cycling_filter()
    elif network_type == 'driving+psv':
        return driving_filter(exclude_public_service_vehicle_paths=False)


def driving_filter(exclude_public_service_vehicle_paths=True):
    """
    Driving filters for different tags (almost) as in OSMnx for 'drive+service'.

    Filter out un-drivable roads, private ways, and
    anything specifying motor=no. also filter out any non-service roads that
    are tagged as providing parking, private, or emergency-access
    services.

    If 'exclude_public_service_vehicle_paths' == False, also paths that are only accessible
    for public transport vehicles are included.

    Applied filters:
        '["area"!~"yes"]["highway"!~"cycleway|footway|path|pedestrian|steps|track|corridor|'
        'elevator|escalator|proposed|construction|bridleway|abandoned|platform|raceway"]'
        '["motor_vehicle"!~"no"]["motorcar"!~"no"]{}'
        '["service"!~"parking|parking_aisle|private|emergency_access"]'

    """
    drive_filter = dict(
        area=['yes'],
        highway=['cycleway', 'footway', 'path', 'pedestrian', 'steps', 'track',
                 'corridor', 'elevator', 'escalator', 'proposed', 'construction',
                 'bridleway', 'abandoned', 'platform', 'raceway'],
        motor_vehicle=['no'],
        motorcar=['no'],
        service=['parking', 'parking_aisle', 'private',
                 'emergency_access']
    )

    if exclude_public_service_vehicle_paths:
        drive_filter['psv'] = ['yes']

    return drive_filter

def walking_filter():
    """
    Walking filters for different tags as in OSMnx for 'walk'.

    Filter out cycle ways, motor ways, private ways, and anything
    specifying foot=no. allow service roads, permitting things like parking
    lot lanes, alleys, etc that you *can* walk on even if they're not exactly
    pleasant walks. some cycleways may allow pedestrians, but this filter ignores
    such cycleways.

    Applied filters:
    '["area"!~"yes"]["highway"!~"cycleway|motor|proposed|construction|abandoned|
      platform|raceway|motorway|motorway_link"]'

    '["foot"!~"no"]["service"!~"private"]'


    """
    return dict(
        area=['yes'],
        highway=['cycleway', 'motor', 'proposed', 'construction', 'abandoned',
                 'platform', 'raceway', 'motorway', 'motorway_link'],
        foot=['no'],
        service=['private']
    )


def cycling_filter():
    """
    Cycling filters for different tags as in OSMnx for 'bike'.

    Filter out foot ways, motor ways, private ways, and anything
    specifying biking=no.

    Applied filters:
        '["area"!~"yes"]["highway"!~"footway|steps|corridor|elevator|escalator|motor|proposed|'
        'construction|abandoned|platform|raceway"]'
        '["bicycle"!~"no"]["service"!~"private"

    """
    return dict(
        area=['yes'],
        highway=['footway', 'steps', 'corridor', 'elevator', 'escalator', 'motor', 'proposed',
                 'construction', 'abandoned', 'platform', 'raceway', 'motorway', 'motorway_link'],
        bicycle=['no'],
        service=['private']
    )



from pyrosm.data_filter cimport filter_osm_records, filter_node_indices, \
    filter_array_dict_by_indices_or_mask, filter_relation_indices
from pyrosm._arrays cimport convert_to_arrays_and_drop_empty, convert_way_records_to_lists, concatenate_dicts_of_arrays
from pyrosm.tagparser cimport explode_tag_array
import numpy as np

cdef get_data_filter_and_osm_keys(custom_filter):
    osm_keys = []
    if isinstance(custom_filter, dict):
        data_filter = {}
        for key, tags in custom_filter.items():
            if not isinstance(key, str):
                raise ValueError(f"OSM key of the 'custom_filter' "
                                 f"should be text. "
                                 f"Got: {key} with type {type(key)}.")
            osm_keys.append(key)

            # # If osm-key has been defined as True,
            # # all tags should be kept, hence, no further filtering
            # # should be done.
            # if tags == True:
            #     continue

            if isinstance(tags, list):
                # Check that all values used for filtering are text (except None/True).
                # Even numeric data should be passed as string.
                for tag in tags:
                    if tag is None:
                        continue
                    if tag is True:
                        continue
                    elif not isinstance(tag, str):
                        raise ValueError(f"OSM tag of the 'custom_filter' "
                                         f"should be text. "
                                         f"Got: {tag} with type {type(tag)}.")
            elif isinstance(tags, str):
                tags = [tags]

            data_filter[key] = tags
        return data_filter, osm_keys
    else:
        raise ValueError(f"'custom_filter' should be a Python dictionary. "
                         f"Got {custom_filter} with type {type(custom_filter)}.")


cdef get_relation_arrays(relations, osm_keys, data_filter, filter_type):
    # Combine all blocks
    relations = concatenate_dicts_of_arrays(relations)
    # Get indices for MultiPolygons that also passes data_filter
    indices = filter_relation_indices(relations, osm_keys, data_filter, filter_type)
    # If no building relations were found, return None
    if len(indices) == 0:
        return None
    # Otherwise, filter the data accordingly
    return filter_array_dict_by_indices_or_mask(relations, indices)

cdef separate_relation_ways(way_records, relation_way_ids):
    cdef int i, n = len(way_records)
    # Create lookup dict
    relation_ids = dict.fromkeys(relation_way_ids, None)
    normal_ways = []
    relation_ways = []
    for i in range(0, n):
        way = way_records[i]
        try:
            # Check if way is part of relation
            relation_ids[way["id"]]
            relation_ways.append(way)
        except Exception as e:
            normal_ways.append(way)
    return normal_ways, relation_ways

cdef get_way_arrays(way_records, relation_way_ids, osm_keys, tags_as_columns, data_filter, filter_type):
    # Get all ways including the ones associated with relations
    ways = filter_osm_records(way_records,
                      data_filter,
                      osm_keys,
                      relation_way_ids,
                      filter_type)

    # If there is not data in the area, do not continue
    # (return None for ways and relations)
    if len(ways) == 0:
        return None, None

    relation_arrays = None
    if relation_way_ids is not None:
        # Separate ways that are part of a relation
        ways, relation_ways = separate_relation_ways(ways, relation_way_ids)
        if len(relation_ways) > 0:
            relation_ways = convert_way_records_to_lists(relation_ways, tags_as_columns)
            relation_arrays = convert_to_arrays_and_drop_empty(relation_ways)

    # Process separated ways
    if len(ways) > 0:
        ways = convert_way_records_to_lists(ways, tags_as_columns)
        way_arrays = convert_to_arrays_and_drop_empty(ways)
    else:
        way_arrays = None

    return way_arrays, relation_arrays

cpdef _get_osm_ways_and_relations(way_records, relations, osm_keys, tags_as_columns, data_filter, filter_type):
    return get_osm_ways_and_relations(way_records, relations, osm_keys, tags_as_columns, data_filter, filter_type)

cdef get_osm_ways_and_relations(way_records, relations, osm_keys, tags_as_columns, data_filter, filter_type):

    # Tags that should always be kept
    tags_as_columns += ["id", "nodes", "timestamp", "version"]

    # If any way records weren't passed in, cannot parse anything
    if way_records is None:
        return None, None, None

    # Get relations for specified OSM keys (one or multiple)
    if relations is not None:
        filtered_relations = get_relation_arrays(relations, osm_keys, data_filter, filter_type)

        # Get all way-ids that are associated with relations
        relation_way_ids = None
        if filtered_relations is not None:
            members = concatenate_dicts_of_arrays(filtered_relations['members'])
            relation_way_ids = members["member_id"]

            if type(relation_way_ids) not in [list, np.ndarray]:
                raise ValueError("'relation_way_ids' should be a list or an array.")
    else:
        relation_way_ids = None
        filtered_relations = None

    # Get ways (separately as "normal" ways and relation_ways)
    ways, relation_ways = get_way_arrays(way_records,
                                         relation_way_ids,
                                         osm_keys,
                                         tags_as_columns,
                                         data_filter,
                                         filter_type)

    # If relation ways could not be parsed, also relations should be returned as None
    if relation_ways is None:
        filtered_relations = None

    # If there weren't any ways or relations return None
    if ways is None and relation_ways is None:
        return None, None, None

    return ways, relation_ways, filtered_relations

cdef get_osm_nodes(node_arrays, osm_keys, tags_as_columns, data_filter, filter_type):
    # Get indices for Nodes that passes data_filter
    indices = filter_node_indices(node_arrays, osm_keys, data_filter, filter_type)
    # If no nodes were found, return None
    if len(indices) == 0:
        return None
    # Otherwise, filter the data accordingly
    filtered_nodes = filter_array_dict_by_indices_or_mask(node_arrays, indices)
    # Explode tags and update
    tags = explode_tag_array(filtered_nodes["tags"], tags_as_columns)
    for k, v in tags.items():
        if list(set(v)) != [None]:
            filtered_nodes[k] = v
    return filtered_nodes

cdef _get_osm_data(node_arrays, way_records, relations, tags_as_columns, data_filter, filter_type, osm_keys):
    if osm_keys is None:
        # Convert filter to appropriate form and parse keys
        data_filter, osm_keys = get_data_filter_and_osm_keys(data_filter)

    if node_arrays is not None:
        # Get nodes
        node_arrays = get_osm_nodes(node_arrays, osm_keys, tags_as_columns, data_filter, filter_type)

    # Parse ways and relations
    ways, relation_ways, filtered_relations = get_osm_ways_and_relations(way_records, relations, osm_keys, tags_as_columns, data_filter, filter_type)
    return node_arrays, ways, relation_ways, filtered_relations

cpdef get_osm_data(node_arrays, way_records, relations, tags_as_columns, data_filter, filter_type, osm_keys=None):
    return _get_osm_data(node_arrays, way_records, relations, tags_as_columns, data_filter, filter_type, osm_keys)



cdef delta_decode_latitude(data, node_granularity, lat_offset)
cdef delta_decode_longitude(data, node_granularity, lon_offset)
cdef delta_decode_id(data)
cdef delta_decode_timestamp(data, timestamp_granularity)
cdef delta_decode_changeset(data)
cdef delta_encode_latitude(lat_array, node_granularity, lat_offset)
cdef delta_encode_longitude(lon_array, node_granularity, lon_offset)



cpdef get_osm_data(node_arrays, way_records, relations, tags_as_columns, data_filter, filter_type, osm_keys=*)
# For debugging purposes
cpdef _get_osm_ways_and_relations(way_records, relations, osm_keys, tags_as_columns, data_filter, filter_type)


from pyrosm.data_manager import get_osm_data
from pyrosm.frames import prepare_geodataframe
from pyrosm.utils import validate_custom_filter
import geopandas as gpd
import warnings


def get_boundary_data(node_coordinates, way_records, relations,
                      tags_as_columns, custom_filter,
                      boundary_type, name, bounding_box):

    if boundary_type == "all":
        boundary_type = True
    else:
        boundary_type = [boundary_type]

    # If custom_filter has not been defined, initialize with default
    if custom_filter is None:
        custom_filter = {"boundary": boundary_type}

    if "boundary" not in custom_filter.keys():
        custom_filter["boundary"] = True

    # Check that the custom filter is in correct format
    custom_filter = validate_custom_filter(custom_filter)

    # Call signature for fetching buildings
    nodes, ways, relation_ways, relations = get_osm_data(node_arrays=None,
                                                         way_records=way_records,
                                                         relations=relations,
                                                         tags_as_columns=tags_as_columns,
                                                         data_filter=custom_filter,
                                                         filter_type="keep",
                                                         osm_keys=None
                                                         )

    # If there weren't any data, return empty GeoDataFrame
    if nodes is None and ways is None and relations is None:
        warnings.warn("Could not find any boundaries for given area.",
                      UserWarning,
                      stacklevel=2)
        return None

    # Prepare GeoDataFrame
    gdf = prepare_geodataframe(nodes, node_coordinates, ways,
                               relations, relation_ways,
                               tags_as_columns, bounding_box)

    if gdf is None:
        return None

    # Filter by name
    # (use Pandas for filtering, which allows using 'contains' more easily)
    if name is not None:
        if "name" not in gdf.columns:
            raise ValueError("Could not filter by name from given area. "
                             "Any of the OSM elements did not have a name tag.")
        gdf = gdf.dropna(subset=["name"])
        gdf = gdf.loc[gdf["name"].str.contains(name)].reset_index(drop=True).copy()

    return gdf



# BASE
URL = "http://download.geofabrik.de/"
suffix = "-latest.osm.pbf"
africa_url = "africa/"
asia_url = "asia/"
australia_oceania_url = "australia-oceania/"
central_america_url = "central-america/"
europe_url = "europe/"
north_america_url = "north-america/"
south_america_url = "south-america/"

baden_wuerttemberg_url = "europe/germany/baden-wuerttemberg/"
bayern_url = "europe/germany/bayern/"
brazil_url = "south-america/brazil/"
canada_url = "north-america/canada/"
england_url = "europe/great-britain/england/"
france_url = "europe/france/"
gb_url = "europe/great-britain/"
germany_url = "europe/germany/"
italy_url = "europe/italy/"
japan_url = "asia/japan/"
netherlands_url = "europe/netherlands/"
nordrhein_wesfalen_url = "europe/germany/nordrhein-westfalen/"
poland_url = "europe/poland/"
russia_url = "russia/"
usa_url = "north-america/us/"


class USA:
    # State level data sources
    regions = [
        "alabama",
        "alaska",
        "arizona",
        "arkansas",
        "colorado",
        "connecticut",
        "delaware",
        "district_of_columbia",
        "florida",
        "georgia",
        "hawaii",
        "idaho",
        "illinois",
        "indiana",
        "iowa",
        "kansas",
        "kentucky",
        "louisiana",
        "maine",
        "maryland",
        "massachusetts",
        "michigan",
        "minnesota",
        "mississippi",
        "missouri",
        "montana",
        "nebraska",
        "nevada",
        "new_hampshire",
        "new_mexico",
        "new_york",
        "new_jersey",
        "north_carolina",
        "north_dakota",
        "ohio",
        "oklahoma",
        "oregon",
        "pennsylvania",
        "puerto_rico",
        "rhode_island",
        "south_carolina",
        "south_dakota",
        "tennessee",
        "texas",
        "utah",
        "vermont",
        "virginia",
        "washington",
        "west_virginia",
        "wisconsin",
        "wyoming",
    ]

    available = regions + ["southern_california", "northern_california"]
    available.sort()

    country = {"name": "us" + suffix, "url": URL + north_america_url + "us" + suffix}

    # Create data sources
    _sources = {
        region: {"name": region.replace("_", "-") + suffix, "url": URL + usa_url + region.replace("_", "-") + suffix}
        for region in regions
    }

    # Add California separately
    _sources["southern_california"] = {
        "name": "socal" + suffix,
        "url": URL + "north-america/us/california/socal-latest.osm.pbf",
    }
    _sources["northern_california"] = {
        "name": "norcal" + suffix,
        "url": URL + "north-america/us/california/norcal-latest.osm.pbf",
    }
    __dict__ = _sources

    def __getattr__(self, name):
        return self.__dict__[name]

    def __call__(self):
        return self.available


class France:
    regions = [
        "alsace",
        "aquitaine",
        "auvergne",
        "basse_normandie",
        "bourgogne",
        "bretagne",
        "centre",
        "champagne_ardenne",
        "corse",
        "franche_comte",
        "guadeloupe",
        "guyane",
        "haute_normandie",
        "ile_de_france",
        "languedoc_roussillon",
        "limousin",
        "lorraine",
        "martinique",
        "mayotte",
        "midi_pyrenees",
        "nord_pas_de_calais",
        "pays_de_la_loire",
        "picardie",
        "poitou_charentes",
        "provence_alpes_cote_d_azur",
        "reunion",
        "rhone_alpes",
    ]

    available = regions
    available.sort()

    country = {"name": "france" + suffix, "url": URL + europe_url + "france" + suffix}

    # Create data sources
    _sources = {
        region: {"name": region.replace("_", "-") + suffix, "url": URL + france_url + region.replace("_", "-") + suffix}
        for region in regions
    }

    __dict__ = _sources

    def __getattr__(self, name):
        return self.__dict__[name]

    def __call__(self):
        return self.available


class England:
    regions = [
        "bedfordshire",
        "berkshire",
        "bristol",
        "buckinghamshire",
        "cambridgeshire",
        "cheshire",
        "cornwall",
        "cumbria",
        "derbyshire",
        "devon",
        "dorset",
        "durham",
        "east_sussex",
        "east_yorkshire_with_hull",
        "essex",
        "gloucestershire",
        "greater_london",
        "greater_manchester",
        "hampshire",
        "herefordshire",
        "hertfordshire",
        "isle_of_wight",
        "kent",
        "lancashire",
        "leicestershire",
        "lincolnshire",
        "merseyside",
        "norfolk",
        "north_yorkshire",
        "northamptonshire",
        "northumberland",
        "nottinghamshire",
        "oxfordshire",
        "rutland",
        "shropshire",
        "somerset",
        "south_yorkshire",
        "staffordshire",
        "suffolk",
        "surrey",
        "tyne_and_wear",
        "warwickshire",
        "west_midlands",
        "west_sussex",
        "west_yorkshire",
        "wiltshire",
        "worcestershire",
    ]

    available = regions
    available.sort()

    # Create data sources
    _sources = {
        region: {"name": region.replace("_", "-") + suffix,
                 "url": URL + england_url + region.replace("_", "-") + suffix}
        for region in regions
    }

    __dict__ = _sources

    def __getattr__(self, name):
        return self.__dict__[name]

    def __call__(self):
        return self.available


class GreatBritain:
    regions = ["england", "scotland", "wales"]
    england = England()

    available = regions + england.available
    available.sort()

    country = {"name": "great-britain" + suffix, "url": URL + europe_url + "great-britain" + suffix}

    # Create data sources
    _sources = {
        region: {"name": region.replace("_", "-") + suffix, "url": URL + gb_url + region.replace("_", "-") + suffix}
        for region in regions
    }

    for region in england.available:
        _sources[region] = {"name": region.replace("_", "-") + suffix,
                            "url": URL + england_url + region.replace("_", "-") + suffix}

    __dict__ = _sources

    def __getattr__(self, name):
        return self.__dict__[name]

    def __call__(self):
        return self.available


class Italy:
    regions = ["centro", "isole", "nord_est", "nord_ovest", "sud"]

    available = regions
    available.sort()

    country = {"name": "italy" + suffix, "url": URL + europe_url + "italy" + suffix}

    # Create data sources
    _sources = {
        region: {"name": region.replace("_", "-") + suffix, "url": URL + italy_url + region.replace("_", "-") + suffix}
        for region in regions
    }

    __dict__ = _sources

    def __getattr__(self, name):
        return self.__dict__[name]

    def __call__(self):
        return self.available


class Russia:
    regions = ["central_fed_district",
               "crimean_fed_district",
               "far_eastern_fed_district",
               "kaliningrad",
               "north_caucasus_fed_district",
               "northwestern_fed_district",
               "siberian_fed_district",
               "south_fed_district",
               "ural_fed_district",
               "volga_fed_district"]

    available = regions
    available.sort()

    country = {"name": "russia" + suffix, "url": URL + russia_url + "russia" + suffix}

    # Create data sources
    _sources = {
        region: {"name": region.replace("_", "-") + suffix, "url": URL + russia_url + region.replace("_", "-") + suffix}
        for region in regions
    }

    __dict__ = _sources

    def __getattr__(self, name):
        return self.__dict__[name]

    def __call__(self):
        return self.available


class Poland:
    regions = ["dolnoslaskie",
               "kujawsko_pomorskie",
               "lodzkie",
               "lubelskie",
               "lubuskie",
               "malopolskie",
               "mazowieckie",
               "opolskie",
               "podkarpackie",
               "podlaskie",
               "pomorskie",
               "slaskie",
               "swietokrzyskie",
               "warminsko_mazurskie",
               "wielkopolskie",
               "zachodniopomorskie"
               ]

    available = regions
    available.sort()

    country = {"name": "poland" + suffix, "url": URL + europe_url + "poland" + suffix}

    # Create data sources
    _sources = {
        region: {"name": region.replace("_", "-") + suffix, "url": URL + poland_url + region.replace("_", "-") + suffix}
        for region in regions
    }

    __dict__ = _sources

    def __getattr__(self, name):
        return self.__dict__[name]

    def __call__(self):
        return self.available


class BadenWuerttemberg:
    regions = ["freiburg_regbez",
               "karlsruhe_regbez",
               "stuttgart_regbez",
               "tuebingen_regbez"]

    available = regions
    available.sort()

    # Create data sources
    _sources = {
        region: {"name": region.replace("_", "-") + suffix,
                 "url": URL + baden_wuerttemberg_url + region.replace("_", "-") + suffix}
        for region in regions
    }

    __dict__ = _sources

    def __getattr__(self, name):
        return self.__dict__[name]

    def __call__(self):
        return self.available


class NordrheinWestfalen:
    regions = ["arnsberg_regbez",
               "detmold_regbez",
               "duesseldorf_regbez",
               "koeln_regbez",
               "muenster_regbez"]

    available = regions
    available.sort()

    # Create data sources
    _sources = {
        region: {"name": region.replace("_", "-") + suffix,
                 "url": URL + nordrhein_wesfalen_url + region.replace("_", "-") + suffix}
        for region in regions
    }

    __dict__ = _sources

    def __getattr__(self, name):
        return self.__dict__[name]

    def __call__(self):
        return self.available


class Bayern:
    regions = ["mittelfranken",
               "niederbayern",
               "oberbayern",
               "oberfranken",
               "oberpfalz",
               "schwaben",
               "unterfranken"]

    available = regions
    available.sort()

    # Create data sources
    _sources = {
        region: {"name": region.replace("_", "-") + suffix,
                 "url": URL + bayern_url + region.replace("_", "-") + suffix}
        for region in regions
    }

    __dict__ = _sources

    def __getattr__(self, name):
        return self.__dict__[name]

    def __call__(self):
        return self.available


class Germany:
    regions = ["baden_wuerttemberg",
               "bayern",
               "berlin",
               "brandenburg",
               "bremen",
               "hamburg",
               "hessen",
               "mecklenburg_vorpommern",
               "niedersachsen",
               "nordrhein_westfalen",
               "rheinland_pfalz",
               "saarland",
               "sachsen_anhalt",
               "sachsen",
               "schleswig_holstein",
               "thueringen"]

    baden_wuerttemberg = BadenWuerttemberg()
    bayern = Bayern()
    nordrhein_westfalen = NordrheinWestfalen()

    available = regions + bayern.available + baden_wuerttemberg.available \
                + nordrhein_westfalen.available
    available.sort()

    country = {"name": "germany" + suffix, "url": URL + europe_url + "germany" + suffix}

    # Create data sources
    _sources = {
        region: {"name": region.replace("_", "-") + suffix,
                 "url": URL + germany_url + region.replace("_", "-") + suffix}
        for region in regions
    }

    for region in nordrhein_westfalen.available:
        _sources[region] = {"name": region.replace("_", "-") + suffix,
                            "url": URL + nordrhein_wesfalen_url + region.replace("_", "-") + suffix}

    for region in baden_wuerttemberg.available:
        _sources[region] = {"name": region.replace("_", "-") + suffix,
                            "url": URL + baden_wuerttemberg_url + region.replace("_", "-") + suffix}

    for region in bayern.available:
        _sources[region] = {"name": region.replace("_", "-") + suffix,
                            "url": URL + bayern_url + region.replace("_", "-") + suffix}

    __dict__ = _sources

    def __getattr__(self, name):
        return self.__dict__[name]

    def __call__(self):
        return self.available


class Netherlands:
    regions = ["drenthe",
               "flevoland",
               "friesland",
               "gelderland",
               "groningen",
               "limburg",
               "noord_brabant",
               "noord_holland",
               "overijssel",
               "utrecht",
               "zeeland",
               "zuid_holland"]
    available = regions
    available.sort()

    country = {"name": "netherlands" + suffix, "url": URL + europe_url + "netherlands" + suffix}

    # Create data sources
    _sources = {
        region: {"name": region.replace("_", "-") + suffix,
                 "url": URL + netherlands_url + region.replace("_", "-") + suffix}
        for region in regions
    }

    __dict__ = _sources

    def __getattr__(self, name):
        return self.__dict__[name]

    def __call__(self):
        return self.available


class Canada:
    regions = ["alberta",
               "british_columbia",
               "manitoba",
               "new_brunswick",
               "newfoundland_and_labrador",
               "northwest_territories",
               "nova_scotia",
               "nunavut",
               "ontario",
               "prince_edward_island",
               "quebec",
               "saskatchewan",
               "yukon"]
    available = regions
    available.sort()

    country = {"name": "canada" + suffix, "url": URL + north_america_url + "canada" + suffix}

    # Create data sources
    _sources = {
        region: {"name": region.replace("_", "-") + suffix, "url": URL + canada_url + region.replace("_", "-") + suffix}
        for region in regions
    }

    __dict__ = _sources

    def __getattr__(self, name):
        return self.__dict__[name]

    def __call__(self):
        return self.available


class Brazil:
    regions = ["centro_oeste", "nordeste", "norte", "sudeste", "sul"]
    available = regions
    available.sort()

    country = {"name": "brazil" + suffix, "url": URL + south_america_url + "brazil" + suffix}

    # Create data sources
    _sources = {
        region: {"name": region.replace("_", "-") + suffix, "url": URL + brazil_url + region.replace("_", "-") + suffix}
        for region in regions
    }

    __dict__ = _sources

    def __getattr__(self, name):
        return self.__dict__[name]

    def __call__(self):
        return self.available


class Japan:
    regions = ["chubu",
               "chugoku",
               "hokkaido",
               "kansai",
               "kanto",
               "kyushu",
               "shikoku",
               "tohoku"]

    available = regions
    available.sort()

    country = {"name": "japan" + suffix, "url": URL + asia_url + "japan" + suffix}

    # Create data sources
    _sources = {
        region: {"name": region.replace("_", "-") + suffix, "url": URL + japan_url + region + suffix}
        for region in regions
    }

    __dict__ = _sources

    def __getattr__(self, name):
        return self.__dict__[name]

    def __call__(self):
        return self.available


class AustraliaOceania:
    regions = ["australia",
               "cook_islands",
               "fiji",
               "kiribati",
               "marshall_islands",
               "micronesia",
               "nauru",
               "new_caledonia",
               "new_zealand",
               "niue",
               "palau",
               "papua_new_guinea",
               "samoa",
               "solomon_islands",
               "tonga",
               "tuvalu",
               "vanuatu"]

    available = regions
    available.sort()

    continent = {"name": "australia-oceania" + suffix, "url": URL + "australia-oceania" + suffix}

    # Create data sources
    _sources = {
        region: {"name": region.replace("_", "-") + suffix,
                 "url": URL + australia_oceania_url + region.replace("_", "-") + suffix}
        for region in regions
    }

    __dict__ = _sources

    def __getattr__(self, name):
        return self.__dict__[name]

    def __call__(self):
        return self.available


class NorthAmerica:
    regions = ["canada",
               "greenland",
               "mexico",
               "usa",
               "us_midwest",
               "us_northeast",
               "us_pacific",
               "us_south",
               "us_west"]

    usa = USA()
    canada = Canada()

    available = regions
    available.sort()

    continent = {"name": "north-america" + suffix, "url": URL + "north-america" + suffix}

    # Create data sources
    _sources = {
        region: {"name": region.replace("_", "-") + suffix,
                 "url": URL + north_america_url + region.replace("_", "-") + suffix}
        for region in regions if region != "usa"
    }
    # USA is "us" in GeoFabrik
    _sources["usa"] = {"name": "us" + suffix,
                       "url": URL + north_america_url + "us" + suffix}

    __dict__ = _sources

    def __getattr__(self, name):
        return self.__dict__[name]

    def __call__(self):
        return self.available


class SouthAmerica:
    regions = ["argentina",
               "bolivia",
               "brazil",
               "chile",
               "colombia",
               "ecuador",
               "paraguay",
               "peru",
               "suriname",
               "uruguay",
               "venezuela"]

    brazil = Brazil()

    available = regions
    available.sort()
    available.sort()

    continent = {"name": "south-america" + suffix, "url": URL + "south-america" + suffix}

    # Create data sources
    _sources = {
        region: {"name": region.replace("_", "-") + suffix, "url": URL + south_america_url + region + suffix}
        for region in regions
    }

    __dict__ = _sources

    def __getattr__(self, name):
        return self.__dict__[name]

    def __call__(self):
        return self.available


class CentralAmerica:
    regions = ["bahamas",
               "belize",
               "cuba",
               "guatemala",
               "haiti_and_domrep",
               "jamaica",
               "nicaragua"]

    available = regions
    available.sort()

    continent = {"name": "central-america" + suffix, "url": URL + "central-america" + suffix}

    # Create data sources
    _sources = {
        region: {"name": region.replace("_", "-") + suffix,
                 "url": URL + central_america_url + region.replace("_", "-") + suffix}
        for region in regions
    }

    __dict__ = _sources

    def __getattr__(self, name):
        return self.__dict__[name]

    def __call__(self):
        return self.available


class Europe:
    # Country specific subregions
    france = France()
    great_britain = GreatBritain()
    italy = Italy()
    russia = Russia()
    poland = Poland()
    germany = Germany()
    netherlands = Netherlands()

    regions = [
        "albania",
        "andorra",
        "austria",
        "azores",
        "belarus",
        "belgium",
        "bosnia_herzegovina",
        "bulgaria",
        "croatia",
        "cyprus",
        "czech_republic",
        "denmark",
        "estonia",
        "faroe_islands",
        "finland",
        "france",
        "georgia",
        "germany",
        "great_britain",
        "greece",
        "hungary",
        "iceland",
        "ireland_and_northern_ireland",
        "isle_of_man",
        "italy",
        "kosovo",
        "latvia",
        "liechtenstein",
        "lithuania",
        "luxembourg",
        "macedonia",
        "malta",
        "moldova",
        "monaco",
        "montenegro",
        "netherlands",
        "norway",
        "poland",
        "portugal",
        "romania",
        "russia",
        "serbia",
        "slovakia",
        "slovenia",
        "spain",
        "sweden",
        "switzerland",
        "turkey",
        "ukraine",
    ]

    available = regions
    available.sort()

    continent = {"name": "europe" + suffix, "url": URL + "europe" + suffix}

    # Create data sources
    _sources = {
        region: {"name": region.replace("_", "-") + suffix, "url": URL + europe_url + region.replace("_", "-") + suffix}
        for region in regions if region != "russia"
    }
    # Russia is separately from Europe
    _sources["russia"] = {"name": "russia" + suffix,
                          "url": URL + "russia" + suffix}

    __dict__ = _sources

    def __getattr__(self, name):
        return self.__dict__[name]

    def __call__(self):
        return self.available


class Africa:
    regions = ["algeria",
               "angola",
               "benin",
               "botswana",
               "burkina_faso",
               "burundi",
               "cameroon",
               "canary_islands",
               "cape_verde",
               "central_african_republic",
               "chad",
               "comores",
               "congo_brazzaville",
               "congo_democratic_republic",
               "djibouti",
               "egypt",
               "equatorial_guinea",
               "eritrea",
               "ethiopia",
               "gabon",
               "ghana",
               "guinea_bissau",
               "guinea",
               "ivory_coast",
               "kenya",
               "lesotho",
               "liberia",
               "libya",
               "madagascar",
               "malawi",
               "mali",
               "mauritania",
               "mauritius",
               "morocco",
               "mozambique",
               "namibia",
               "niger",
               "nigeria",
               "rwanda",
               "saint_helena_ascension_and_tristan_da_cunha",
               "sao_tome_and_principe",
               "senegal_and_gambia",
               "seychelles",
               "sierra_leone",
               "somalia",
               "south_africa_and_lesotho",
               "south_africa",
               "south_sudan",
               "sudan",
               "swaziland",
               "tanzania",
               "togo",
               "tunisia",
               "uganda",
               "zambia",
               "zimbabwe"]

    available = regions
    available.sort()

    continent = {"name": "africa" + suffix, "url": URL + "africa" + suffix}

    # Create data sources
    _sources = {
        region: {"name": region.replace("_", "-") + suffix, "url": URL + africa_url + region.replace("_", "-") + suffix}
        for region in regions
    }

    __dict__ = _sources

    def __getattr__(self, name):
        return self.__dict__[name]

    def __call__(self):
        return self.available


class Asia:
    regions = ["afghanistan",
               "armenia",
               "azerbaijan",
               "bangladesh",
               "bhutan",
               "cambodia",
               "china",
               "gcc_states",
               "india",
               "indonesia",
               "iran",
               "iraq",
               "israel_and_palestine",
               "japan",
               "jordan",
               "kazakhstan",
               "kyrgyzstan",
               "laos",
               "lebanon",
               "malaysia_singapore_brunei",
               "maldives",
               "mongolia",
               "myanmar",
               "nepal",
               "north_korea",
               "pakistan",
               "philippines",
               "south_korea",
               "sri_lanka",
               "syria",
               "taiwan",
               "tajikistan",
               "thailand",
               "turkmenistan",
               "uzbekistan",
               "vietnam",
               "yemen"]

    japan = Japan()

    available = regions
    available.sort()

    continent = {"name": "asia" + suffix, "url": URL + "asia" + suffix}

    # Create data sources
    _sources = {
        region: {"name": region.replace("_", "-") + suffix, "url": URL + asia_url + region.replace("_", "-") + suffix}
        for region in regions
    }

    __dict__ = _sources

    def __getattr__(self, name):
        return self.__dict__[name]

    def __call__(self):
        return self.available


class Antarctica:
    regions = ["antarctica"]
    available = regions
    available.sort()

    continent = {"name": "antarctica" + suffix, "url": URL + "antarctica" + suffix}

    # Create data sources
    _sources = {
        region: {"name": region + suffix, "url": URL + region + suffix}
        for region in regions
    }

    __dict__ = _sources

    def __getattr__(self, name):
        return self.__dict__[name]

    def __call__(self):
        return self.available


class SubRegions:
    def __init__(self):
        self.regions = ["brazil", "canada", "france", "germany", "great_britain", "italy", "japan", "netherlands", "poland", "russia", "usa"]
        available = self.regions
        available.sort()

        self.brazil = Brazil()
        self.canada = Canada()
        self.france = France()
        self.germany = Germany()
        self.great_britain = GreatBritain()
        self.italy = Italy()
        self.japan = Japan()
        self.netherlands = Netherlands()
        self.poland = Poland()
        self.russia = Russia()
        self.usa = USA()

        self.available = {name: self.__dict__[name].available for name in self.regions}

    def __getattr__(self, name):
        return self.__dict__[name]

    def __call__(self):
        return self.available
